# 7 "parsing_cocci/lexer_cocci.mll"
 
open Parser_cocci_menhir
module D = Data
module Ast = Ast_cocci
module Ast0 = Ast0_cocci
module P = Parse_aux
module FC = Flag_parsing_cocci
exception Lexical of string
let tok = Lexing.lexeme

let line = ref 1
let file = ref ""
let logical_line = ref 0

(* ---------------------------------------------------------------------- *)
(* control codes *)

(* Defined in data.ml
type line_type = MINUS | OPTMINUS | PLUS | CONTEXT | OPT
*)

let current_line_type = ref (D.CONTEXT,!line,!logical_line)

let prev_plus = ref false
let line_start = ref 0 (* offset of the beginning of the line *)
let get_current_line_type lexbuf =
  let (c,l,ll) = !current_line_type in
  let lex_start = Lexing.lexeme_start lexbuf in
  let preceeding_spaces =
    if !line_start < 0 then 0 else lex_start - !line_start in
  (*line_start := -1;*)
  prev_plus := (c = D.PLUS) || (c = D.PLUSPLUS);
  (c,l,ll,ll,lex_start,preceeding_spaces,[],[],[],"")
let current_line_started = ref false
let col_zero = ref true

let contextify (c,l,ll,lle,lex_start,preceeding_spaces,bef,aft,pos,ws) =
  (D.CONTEXT,l,ll,lle,lex_start,preceeding_spaces,bef,aft,pos,ws)

let reset_line lexbuf =
  line := !line + 1;
  current_line_type := (D.CONTEXT,!line,!logical_line);
  current_line_started := false;
  (if not(!Data.in_meta || !Data.in_rule_name) then col_zero := true);
  line_start := Lexing.lexeme_start lexbuf + 1

let started_line = ref (-1)

let start_line seen_char =
  current_line_started := true;
  col_zero := false;
  (if seen_char && not(!line = !started_line)
  then
    begin
      started_line := !line;
      logical_line := !logical_line + 1
    end)

let pass_zero _ = col_zero := false

let lexerr s1 s2 = raise (Lexical (Printf.sprintf "%s%s" s1 s2))

let opt_reverse_token token =
  if !FC.interpret_inverted
  then match token with
         D.MINUS        -> D.PLUSPLUS  (* maybe too liberal *)
       | D.OPTMINUS     -> lexerr "cannot invert token ?- (an optional minus line), which is needed for reversing the patch" ""
       | D.PLUS         -> D.MINUS
       | D.PLUSPLUS     -> D.MINUS (* may not be sufficient *)
       | _              -> token
  else token

let add_current_line_type x =
  match (opt_reverse_token x,!current_line_type) with
    (D.MINUS,(D.CONTEXT,ln,lln))  ->
      current_line_type := (D.MINUS,ln,lln)
  | (D.MINUS,(D.OPT,ln,lln))      ->
      current_line_type := (D.OPTMINUS,ln,lln)
  | (D.PLUS,(D.CONTEXT,ln,lln))   ->
      current_line_type := (D.PLUS,ln,lln)
  | (D.PLUSPLUS,(D.CONTEXT,ln,lln))   ->
      current_line_type := (D.PLUSPLUS,ln,lln)
  | (D.OPT,(D.CONTEXT,ln,lln))    ->
      current_line_type := (D.OPT,ln,lln)
  | _ -> lexerr "invalid control character combination" ""

let check_minus_context_linetype s =
  match !current_line_type with
    (D.PLUS,_,_) | (D.PLUSPLUS,_,_) -> lexerr "invalid in a + context: " s
  | _ -> ()

let check_context_linetype s =
  match !current_line_type with
    (D.CONTEXT,_,_) -> ()
  | _ -> lexerr "invalid in a nonempty context: " s

let check_plus_linetype s =
  match !current_line_type with
    (D.PLUS,_,_) | (D.PLUSPLUS,_,_) -> ()
  | _ -> lexerr "invalid in a non + context: " s

let check_arity_context_linetype s =
  match !current_line_type with
    (D.CONTEXT,_,_) | (D.PLUS,_,_) | (D.PLUSPLUS,_,_)
  | (D.OPT,_,_) -> ()
  | _ -> lexerr "invalid in a nonempty context: " s

let check_comment s =
  if not !current_line_started
  then lexerr "+ expected at the beginning of the line" s

let process_include start finish str =
  (match !current_line_type with
    (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
      (try
	let _ = Str.search_forward (Str.regexp "\\.\\.\\.") str start in
	lexerr "... not allowed in + include" ""
      with Not_found -> ())
  | _ -> ());
  String.sub str (start + 1) (finish - start - 1)

(* ---------------------------------------------------------------------- *)
type pm = PATCH | MATCH | UNKNOWN

let pm = ref UNKNOWN

let patch_or_match = function
    PATCH ->
      if not !D.ignore_patch_or_match
      then
	(match !pm with
	  MATCH ->
	    lexerr "- or + not allowed in the first column for a match" ""
	| PATCH -> ()
	| UNKNOWN -> Flag.sgrep_mode2 := false; pm := PATCH)
  | MATCH ->
      if not !D.ignore_patch_or_match
      then
	(match !pm with
	  PATCH -> lexerr "* not allowed in the first column for a patch" ""
	| MATCH -> ()
	| UNKNOWN -> Flag.sgrep_mode2 := true; pm := MATCH)
  | _ -> failwith "unexpected argument"

(* ---------------------------------------------------------------------- *)
(* identifiers, including metavariables *)

let metavariables = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let all_metavariables =
  (Hashtbl.create(100) : (string,(string * (D.clt -> token)) list) Hashtbl.t)

let type_names = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let attr_names = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let declarer_names = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let iterator_names = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let symbol_names = (Hashtbl.create(15) : (string, D.clt -> token) Hashtbl.t)

let rule_names = (Hashtbl.create(100) : (string, unit) Hashtbl.t)

let check_var s linetype =
  let fail _ =
    if (!Data.in_prolog || !Data.in_rule_name) &&
      Str.string_match (Str.regexp "<.*>") s 0
    then TPathIsoFile s
    else
      try (Hashtbl.find metavariables s) linetype
      with Not_found ->
	(try (Hashtbl.find type_names s) linetype
	with Not_found ->
	  (try (Hashtbl.find attr_names s) linetype
	  with Not_found ->
	    (try (Hashtbl.find declarer_names s) linetype
	    with Not_found ->
	      (try (Hashtbl.find iterator_names s) linetype
	      with Not_found ->
		(try (Hashtbl.find symbol_names s) linetype
		with Not_found ->
                  TIdent (s,linetype)))))) in
  if !Data.in_meta || !Data.in_rule_name
  then (try Hashtbl.find rule_names s; TRuleName s with Not_found -> fail())
  else fail()

let id_tokens lexbuf =
  let s = tok lexbuf in
  let linetype = get_current_line_type lexbuf in
  let in_rule_name = !Data.in_rule_name in
  let in_meta = !Data.in_meta && not !Data.saw_struct in
  let in_iso = !Data.in_iso in
  let in_prolog = !Data.in_prolog in
  (if s = "identifer" && in_meta
  then Common.pr2 "Warning: should identifer be identifier?");
  match s with
    "metavariable" when in_meta -> check_arity_context_linetype s; TMetavariable
  | "identifier" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TIdentifier
  | "type" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TType
  | "parameter" when in_meta ->  check_arity_context_linetype s; TParameter
  | "operator" when in_meta ->   check_arity_context_linetype s; TOperator
  | "binary" when in_meta ->   check_arity_context_linetype s; TBinary
  | "assignment" when in_meta ->   check_arity_context_linetype s; TAssignment
  | "constant"  when in_meta ->  check_arity_context_linetype s; TConstant
  | "generated" when in_rule_name && not (!Flag.make_hrule = None) ->
      check_arity_context_linetype s; TGenerated
  | "expression" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TExpression
  | "declaration" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TDeclaration
  | "field" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TField
  | "initialiser" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TInitialiser
  | "initializer" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TInitialiser
  | "idexpression" when in_meta ->
      check_arity_context_linetype s; TIdExpression
  | "statement" when in_meta ->  check_arity_context_linetype s; TStatement
  | "function"  when in_meta ->  check_arity_context_linetype s; TFunction
  | "local" when in_meta ->      check_arity_context_linetype s; TLocal
  | "global" when in_meta ->     check_arity_context_linetype s; TGlobal
  | "list" when in_meta ->       check_arity_context_linetype s; Tlist
  | "fresh" when in_meta ->      check_arity_context_linetype s; TFresh
  | "typedef" when in_meta ->    check_arity_context_linetype s; TTypedef
  | "attribute" when in_meta ->  check_arity_context_linetype s; TAttribute
  | "declarer" when in_meta ->   check_arity_context_linetype s; TDeclarer
  | "iterator" when in_meta ->   check_arity_context_linetype s; TIterator
  | "name" when in_meta ->       check_arity_context_linetype s; TName
  | "position" when in_meta ->   check_arity_context_linetype s; TPosition
  | "comments" when in_meta ->   check_arity_context_linetype s; TComments
  | "format" when in_meta ->     check_arity_context_linetype s; TFormat
  | "analysis" when in_meta ->   check_arity_context_linetype s; TAnalysis
  | "any" when in_meta ->        check_arity_context_linetype s; TPosAny
  | "pure" when in_meta && in_iso ->
      check_arity_context_linetype s; TPure
  | "context" when in_meta && in_iso ->
      check_arity_context_linetype s; TContext
  | "error" when in_meta ->      check_arity_context_linetype s; TError
  | "words" when in_meta ->      check_context_linetype s; TWords
  | "symbol" when in_meta ->     check_arity_context_linetype s; TSymbol

  | "using" when in_rule_name || in_prolog ->  check_context_linetype s; TUsing
  | "virtual" when in_prolog || in_rule_name || in_meta ->
      (* don't want to allow virtual as a rule name *)
      check_context_linetype s; TVirtual
  | "merge" when in_prolog || in_rule_name || in_meta ->
      check_context_linetype s; TMerge
  | "disable" when in_rule_name ->  check_context_linetype s; TDisable
  | "extends" when in_rule_name -> check_context_linetype s; TExtends
  | "depends" when in_rule_name -> check_context_linetype s; TDepends
  | "on" when in_rule_name      -> check_context_linetype s; TOn
  | "ever" when in_rule_name    -> check_context_linetype s; TEver
  | "never" when in_rule_name   -> check_context_linetype s; TNever
  | "file" when in_rule_name    -> check_context_linetype s; TFile
  | "in" when in_rule_name      -> check_context_linetype s; TIn
  (* exists and forall for when are reparsed in parse_cocci.ml *)
  | "exists" when in_rule_name  -> check_context_linetype s; TExists
  | "forall" when in_rule_name  -> check_context_linetype s; TForall
  | "script" when (in_rule_name || in_meta) ->
      check_context_linetype s; TScript (!file, !line)
  | "initialize" when in_rule_name -> check_context_linetype s; TInitialize
  | "finalize" when in_rule_name   -> check_context_linetype s; TFinalize

  | "char" ->       Tchar     linetype
  | "short" ->      Tshort    linetype
  | "int" ->        Tint      linetype
  | "double" ->     Tdouble   linetype
  | "float" ->      Tfloat    linetype
  | "complex" ->    Tcomplex  linetype
  | "long" ->       Tlong     linetype
  | "void" ->       Tvoid     linetype
  | "size_t" ->     Tsize_t   linetype
  | "ssize_t" ->    Tssize_t  linetype
  | "ptrdiff_t" ->  Tptrdiff_t linetype
  (* in_meta is only for the first keyword; drop it now to allow any type
     name *)
  | "struct" ->     Data.saw_struct := true; Tstruct   linetype
  | "union" ->      Data.saw_struct := true; Tunion    linetype
  | "enum" ->       Data.saw_struct := true; Tenum     linetype
  | "unsigned" ->   Tunsigned linetype
  | "signed" ->     Tsigned   linetype
  | "decimal" when !Flag.ibm -> Tdecimal linetype
  | "EXEC" when !Flag.ibm -> Texec linetype

  | "auto"  ->
      let open Flag in
      begin match !c_plus_plus with
        On None ->
          raise (
            Semantic_cocci.Semantic (
              "auto has different meaning in different versions of
              C++. Please specify a version using --c++=<version>"))
      | On (Some i) ->
          if i >= 2011
          then TautoType linetype
          else Tauto linetype
      | Off -> Tauto linetype end
  | "register" ->   Tregister linetype
  | "extern" ->     Textern   linetype
  | "static" ->     Tstatic   linetype
  | "inline" ->     Tinline   linetype
  | "typedef" ->    Ttypedef  linetype

  | "const" ->      Tconst    linetype
  | "volatile" ->   Tvolatile linetype

  | "if" ->         TIf       linetype
  | "else" ->       TElse     linetype
  | "while" ->      TWhile    linetype
  | "do" ->         TDo       linetype
  | "for" ->        TFor      linetype
  | "switch" ->     TSwitch   linetype
  | "case" ->       TCase     linetype
  | "default" ->    TDefault  linetype
  | "return" ->     TReturn   linetype
  | "break" ->      TBreak    linetype
  | "continue" ->   TContinue linetype
  | "goto" ->       TGoto     linetype

  | "sizeof" ->     TSizeof   linetype
  | "typeof" ->     TTypeof   linetype

  | "__attribute__" -> TAttr_ linetype

  | "Expression"       when !Data.in_iso -> TIsoExpression
  | "ArgExpression"    when !Data.in_iso -> TIsoArgExpression
  | "TestExpression"   when !Data.in_iso -> TIsoTestExpression
  | "ToTestExpression" when !Data.in_iso -> TIsoToTestExpression
  | "Statement"        when !Data.in_iso -> TIsoStatement
  | "Declaration"      when !Data.in_iso -> TIsoDeclaration
  | "Type"             when !Data.in_iso -> TIsoType
  | "TopLevel"         when !Data.in_iso -> TIsoTopLevel

  | "_" when !Data.in_meta -> TUnderscore

  | s -> check_var s linetype

let mkassign op lexbuf =
  TOpAssign (op, (get_current_line_type lexbuf))

let spinit _ = (* per semantic patch *)
  Hashtbl.clear Data.non_local_script_constraints;
  D.type_names := [];
  D.attr_names := [];
  D.iterator_names := [];
  D.declarer_names := []

let init _ = (* per file, first .cocci then iso *)
  line := 1;
  logical_line := 0;
  prev_plus := false;
  line_start := 0;
  current_line_started := false;
  current_line_type := (D.CONTEXT,0,0);
  col_zero := true;
  pm := UNKNOWN;
  Data.in_rule_name := false;
  Data.in_meta := false;
  Data.in_prolog := false;
  Data.saw_struct := false;
  Data.inheritable_positions := [];
  Hashtbl.clear all_metavariables;
  Hashtbl.clear Data.all_metadecls;
  Hashtbl.clear metavariables;
  Hashtbl.clear type_names;
  Hashtbl.clear attr_names;
  Hashtbl.clear rule_names;
  Hashtbl.clear iterator_names;
  Hashtbl.clear declarer_names;
  Hashtbl.clear symbol_names;
  let get_name (_,x) = x in
  Data.add_meta_meta :=
    (fun name cstr pure ->
      let fn clt = TMeta(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_id_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaId(name,constraints,Ast.NoVal,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_virt_id_meta_found :=
    (fun name vl ->
      let fn clt = TIdent(vl,clt) in
      Hashtbl.replace metavariables name fn);
  Data.add_virt_id_meta_not_found :=
    (fun name pure ->
      let fn clt = TMetaId(name,Ast.CstrTrue,Ast.NoVal,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_fresh_id_meta :=
    (fun name seed ->
      let fn clt = TMetaId(name,Ast.CstrTrue,seed,Ast0.Impure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_type_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaType(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_init_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaInit(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_initlist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaInitList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_param_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaParam(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_paramlist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaParamList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_const_meta :=
    (fun tyopt name constraints pure ->
      let fn clt = TMetaConst(name,constraints,pure,tyopt,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_err_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaErr(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_exp_meta :=
    (fun tyopt name constraints pure bitfield ->
      let fn clt = TMetaExp(name,constraints,pure,tyopt,clt,bitfield) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_idexp_meta :=
    (fun tyopt name constraints pure ->
      let fn clt = TMetaIdExp(name,constraints,pure,tyopt,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_local_idexp_meta :=
    (fun tyopt name constraints pure ->
      let fn clt = TMetaLocalIdExp(name,constraints,pure,tyopt,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_global_idexp_meta :=
    (fun tyopt name constraints pure ->
      let fn clt = TMetaGlobalIdExp(name,constraints,pure,tyopt,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_explist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaExpList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_decl_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaDecl(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_field_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaField(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_field_list_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaFieldList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_fmt_meta :=
    (function name -> function iconstraints ->
      let fn clt = failwith "format metavariable only allowed in a string" in
      Data.format_metavariables :=
	(get_name name,(name,iconstraints)) :: !Data.format_metavariables;
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_fmtlist_meta :=
    (fun name constraints lenname ->
      let fn clt =
	failwith "format list metavariable only allowed in a string" in
      Data.format_list_metavariables :=
	(get_name name,(name,lenname,constraints))
	:: !Data.format_list_metavariables;
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_stm_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaStm(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_stmlist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaStmList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_dparamlist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaDParamList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_func_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaFunc(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_local_func_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaLocalFunc(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_iterator_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaIterator(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_declarer_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaDeclarer(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_pos_meta :=
    (fun name constraints any ->
      let fn ((d,ln,_,_,_,_,_,_,_,_) as clt) =
	(if d = Data.PLUS
	then
	  failwith
	    (Printf.sprintf "%d: positions only allowed in minus code" ln));
	TMetaPos(name,constraints,any,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_com_meta :=
    (fun name constraints ->
      let fn ((d,ln,_,_,_,_,_,_,_,_) as clt) =
	(if d = Data.PLUS
	then
	  failwith
	    (Printf.sprintf "%d: comment variables only allowed in minus code" ln));
	TMetaCom(name,constraints,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_assignOp_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaAssignOp (name, constraints, pure, clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_binaryOp_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaBinaryOp (name, constraints, pure, clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_type_name :=
    (function name ->
      let fn clt = TTypeId(name,clt) in
      (if not (List.mem name !D.type_names)
      then D.type_names := name :: !D.type_names);
      Hashtbl.replace type_names name fn);
  Data.add_attribute :=
    (function name ->
      let fn clt =
	match clt with
	  ((Data.PLUS | Data.PLUSPLUS),_,_,_,_,_,_,_,_,_) ->
	    TDirective (Ast.Space name, clt)
	| _ -> Tattr (name, clt) in
      (if not (List.mem name !D.attr_names)
      then D.attr_names := name :: !D.attr_names);
      Hashtbl.replace attr_names name fn);
  Data.add_attribute_meta :=
    (fun name cstr pure ->
      let fn ((d,ln,_,_,_,_,_,_,_,_) as clt) =
        if (match d with (Data.PLUS | Data.PLUSPLUS) -> true | _ -> false)
        then
          (* TODO support meta attributes in plus code *)
	  failwith
	    (Printf.sprintf
             "%d: meta attributes currently only allowed in context/minus code"
             ln);
        TMetaAttribute(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_declarer_name :=
    (function name ->
      let fn clt = TDeclarerId(name,clt) in
      (if not (List.mem name !D.declarer_names)
      then D.declarer_names := name :: !D.declarer_names);
      Hashtbl.replace declarer_names name fn);
  Data.add_iterator_name :=
    (function name ->
      let fn clt = TIteratorId(name,clt) in
      (if not (List.mem name !D.iterator_names)
      then D.iterator_names := name :: !D.iterator_names);
      Hashtbl.replace iterator_names name fn);
  Data.add_symbol_meta :=
    (function name ->
      let fn clt = TSymId (name,clt) in
      Hashtbl.replace symbol_names name fn);
  Data.init_rule := (function _ -> Hashtbl.clear metavariables);
  Data.install_bindings :=
    (function parent ->
      List.iter (function (name,fn) -> Hashtbl.add metavariables name fn)
	(Hashtbl.find all_metavariables parent))

let post_init _ =
  Stdcompat.Hashtbl.reset all_metavariables;
  Stdcompat.Hashtbl.reset Data.all_metadecls;
  Stdcompat.Hashtbl.reset metavariables;
  Stdcompat.Hashtbl.reset type_names;
  Stdcompat.Hashtbl.reset attr_names;
  Stdcompat.Hashtbl.reset rule_names;
  Stdcompat.Hashtbl.reset iterator_names;
  Stdcompat.Hashtbl.reset declarer_names;
  Stdcompat.Hashtbl.reset symbol_names

(* initialization for each cocci rule *)
let reinit _ =
  Data.format_metavariables := [];
  Data.format_list_metavariables := []

(* the following is needed to properly tokenize include files.  Because an
include file is included after seeing a @, so current_line_started is true.
Current_line_started is not important for parsing the name of a rule, so we
don't have to reset this value to true after parsing an included file. *)
let include_init _ =
  current_line_started := false

let drop_spaces s =
  let len = String.length s in
  let rec loop n =
    if n = len
    then n
    else
      if List.mem (String.get s n) [' ';'\t']
      then loop (n+1)
      else n in
  let start = loop 0 in
  String.sub s start (len - start)

# 611 "parsing_cocci/lexer_cocci.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\153\255\154\255\081\000\119\000\167\255\168\255\192\000\
    \023\001\112\001\201\001\034\002\038\000\067\000\101\000\079\000\
    \218\255\081\000\082\000\085\000\224\255\226\255\229\255\230\255\
    \231\255\232\255\234\255\086\000\192\000\238\255\087\000\240\255\
    \116\000\118\000\140\000\111\002\122\002\209\002\087\000\238\002\
    \150\000\255\255\068\003\205\000\105\000\150\000\041\000\203\000\
    \154\000\057\000\039\000\057\000\054\000\152\000\240\002\175\000\
    \177\255\211\255\073\000\107\000\089\000\076\000\092\000\084\000\
    \188\255\075\003\096\000\105\000\106\000\222\000\146\003\221\003\
    \107\000\108\000\106\000\117\000\052\001\079\004\162\004\185\255\
    \125\000\120\000\115\000\128\000\054\001\020\005\134\005\092\003\
    \241\000\242\002\127\000\121\000\117\000\153\000\154\000\076\003\
    \237\000\004\001\036\001\079\003\080\001\183\255\006\001\182\255\
    \238\000\181\255\072\001\253\001\075\001\116\002\077\001\204\002\
    \094\003\254\001\171\000\186\000\042\001\186\000\220\000\255\001\
    \117\002\218\000\223\000\118\002\004\001\002\001\120\002\007\001\
    \018\001\159\002\250\255\091\001\192\005\023\006\110\006\197\006\
    \028\007\094\001\115\007\202\007\004\008\091\008\149\008\095\001\
    \029\001\236\008\098\001\100\001\067\009\154\009\101\001\241\009\
    \072\010\130\010\217\010\048\011\135\011\078\003\125\001\033\002\
    \127\001\112\001\245\255\243\255\157\001\225\255\113\003\047\004\
    \064\004\142\001\144\001\145\001\162\001\190\001\192\001\246\255\
    \193\001\196\001\244\255\205\255\204\255\156\255\214\255\198\001\
    \228\255\171\002\213\255\201\001\173\002\200\255\004\000\178\255\
    \190\255\233\255\235\255\237\255\194\255\208\255\212\255\199\255\
    \184\001\185\001\206\255\203\255\210\255\155\255\201\255\195\255\
    \209\255\207\255\193\011\024\012\189\001\082\012\169\012\190\001\
    \000\013\087\013\145\013\232\013\165\255\166\255\163\255\164\255\
    \161\255\162\255\065\014\160\255\157\255\029\003\217\001\140\014\
    \074\004\243\004\010\005\167\014\084\004\020\005\237\014\219\001\
    \158\255\008\002\075\015\189\255\155\015\193\015\200\255\201\255\
    \011\016\098\016\187\016\020\017\109\017\003\002\239\001\013\002\
    \124\002\051\003\154\002\017\002\020\002\035\002\240\002\239\255\
    \240\255\096\005\241\002\243\255\244\255\245\255\246\255\247\255\
    \248\255\125\002\015\003\034\002\096\003\255\255\145\005\137\002\
    \203\002\209\255\216\255\252\255\236\255\251\255\235\255\250\255\
    \219\255\242\255\100\006\119\005\186\006\017\007\234\255\185\002\
    \211\255\217\255\218\255\215\255\225\255\195\002\210\255\214\255\
    \222\255\213\255\221\255\212\255\167\017\254\017\210\002\056\018\
    \143\018\033\003\230\018\061\019\119\019\206\019\208\255\198\255\
    \199\255\034\003\037\020\124\020\035\003\211\020\042\021\100\021\
    \187\021\245\021\037\003\228\002\076\022\048\003\051\003\163\022\
    \250\022\061\003\081\023\168\023\226\023\057\024\196\255\197\255\
    \194\255\195\255\146\024\193\255\190\255\064\003\045\003\221\024\
    \107\007\062\009\233\009\248\024\117\007\080\008\062\025\066\003\
    \191\255\056\003\255\255\076\009\252\255\132\025\015\006\194\007\
    \254\255\155\025\253\255\105\003\254\255\228\014\255\255\251\255\
    \193\025\228\008\243\009\253\255\216\025\252\255\216\014\253\255\
    \051\015\254\255\255\255\250\255\254\025\061\010\209\010\252\255\
    \021\026\251\255\220\014\252\255\253\255\254\255\076\003\255\255\
    ";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\097\000\097\000\255\255\255\255\082\000\
    \081\000\081\000\081\000\081\000\053\000\066\000\064\000\040\000\
    \255\255\036\000\059\000\032\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\019\000\102\000\255\255\016\000\255\255\
    \014\000\013\000\058\000\028\000\081\000\081\000\006\000\102\000\
    \033\000\255\255\001\000\255\255\255\255\002\000\255\255\255\255\
    \003\000\255\255\255\255\255\255\255\255\004\000\004\000\004\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\068\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\069\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\071\000\
    \071\000\076\000\255\255\255\255\255\255\255\255\255\255\075\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\255\255\255\255\255\255\255\255\255\255\076\000\
    \076\000\255\255\255\255\076\000\255\255\255\255\076\000\255\255\
    \255\255\076\000\255\255\255\255\255\255\081\000\081\000\007\000\
    \255\255\084\000\255\255\085\000\255\255\083\000\255\255\255\255\
    \255\255\255\255\083\000\255\255\255\255\255\255\085\000\255\255\
    \085\000\255\255\255\255\081\000\081\000\096\000\255\255\008\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\096\000\
    \255\255\057\000\063\000\034\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\039\000\
    \255\255\079\000\255\255\038\000\080\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \062\000\035\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\086\000\255\255\255\255\255\255\086\000\
    \255\255\086\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\081\000\255\255\255\255\097\000\097\000\255\255\
    \096\000\255\255\255\255\097\000\096\000\255\255\097\000\097\000\
    \255\255\097\000\255\255\255\255\064\000\064\000\255\255\255\255\
    \049\000\048\000\048\000\048\000\048\000\066\000\035\000\029\000\
    \028\000\027\000\032\000\025\000\022\000\018\000\031\000\255\255\
    \255\255\014\000\023\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\066\000\006\000\066\000\024\000\255\255\001\000\255\255\
    \002\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\063\000\255\255\063\000\255\255\255\255\017\000\
    \255\255\255\255\255\255\255\255\255\255\026\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\053\000\255\255\255\255\
    \255\255\053\000\255\255\053\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\051\000\255\255\052\000\255\255\
    \050\000\255\255\255\255\255\255\255\255\050\000\255\255\255\255\
    \255\255\052\000\255\255\052\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\048\000\255\255\255\255\064\000\064\000\255\255\
    \063\000\255\255\255\255\064\000\063\000\255\255\064\000\064\000\
    \255\255\255\255\255\255\000\000\255\255\003\000\001\000\001\000\
    \255\255\002\000\255\255\255\255\255\255\001\000\255\255\255\255\
    \004\000\002\000\002\000\255\255\003\000\255\255\255\255\255\255\
    \002\000\255\255\255\255\255\255\005\000\003\000\003\000\255\255\
    \004\000\255\255\255\255\255\255\255\255\255\255\004\000\255\255\
    ";
  Lexing.lex_default =
   "\001\000\000\000\000\000\255\255\255\255\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\000\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\045\000\255\255\255\255\
    \048\000\255\255\255\255\255\255\255\255\255\255\055\000\055\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\088\000\
    \088\000\106\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\102\000\100\000\255\255\100\000\000\000\102\000\000\000\
    \255\255\000\000\106\000\106\000\106\000\106\000\110\000\106\000\
    \106\000\113\000\255\255\255\255\255\255\255\255\255\255\119\000\
    \120\000\255\255\255\255\123\000\255\255\255\255\126\000\255\255\
    \255\255\129\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\185\000\000\000\255\255\188\000\000\000\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\243\000\000\000\255\255\255\255\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \000\000\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \024\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\106\001\000\000\108\001\000\000\255\255\255\255\255\255\
    \000\000\255\255\000\000\116\001\000\000\119\001\000\000\000\000\
    \255\255\255\255\255\255\000\000\255\255\000\000\127\001\000\000\
    \131\001\000\000\000\000\000\000\255\255\255\255\255\255\000\000\
    \255\255\000\000\139\001\000\000\000\000\000\000\255\255\000\000\
    ";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\042\000\041\000\041\000\041\000\041\000\191\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \042\000\030\000\005\000\039\000\007\000\017\000\014\000\006\000\
    \029\000\026\000\019\000\032\000\021\000\033\000\035\000\040\000\
    \004\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\012\000\020\000\034\000\015\000\018\000\031\000\
    \038\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\011\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\010\000\008\000\037\000\
    \008\000\008\000\008\000\025\000\028\000\024\000\013\000\008\000\
    \210\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\009\000\008\000\036\000\
    \008\000\008\000\008\000\023\000\027\000\022\000\016\000\232\000\
    \209\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\207\000\206\000\205\000\204\000\199\000\
    \200\000\201\000\198\000\197\000\189\000\228\000\233\000\130\000\
    \045\000\049\000\050\000\051\000\052\000\229\000\053\000\187\000\
    \255\255\054\000\208\000\183\000\255\255\232\000\230\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\234\000\
    \234\000\186\000\127\000\182\000\184\000\228\000\233\000\172\000\
    \054\000\255\255\173\000\228\000\233\000\229\000\080\000\089\000\
    \056\000\072\000\066\000\229\000\067\000\045\000\230\000\090\000\
    \170\000\169\000\191\000\171\000\230\000\190\000\068\000\231\000\
    \069\000\073\000\196\000\057\000\047\000\074\000\047\000\116\000\
    \075\000\115\000\076\000\228\000\233\000\114\000\081\000\082\000\
    \083\000\084\000\091\000\229\000\007\000\092\000\192\000\070\000\
    \195\000\193\000\093\000\047\000\230\000\047\000\048\000\231\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\131\000\255\255\132\000\094\000\070\000\095\000\
    \002\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\104\000\105\000\124\000\121\000\007\000\
    \120\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\194\000\077\000\046\000\085\000\
    \046\000\119\000\255\255\122\000\103\000\123\000\255\255\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\131\000\255\255\132\000\077\000\255\255\085\000\255\255\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\101\000\125\000\126\000\128\000\008\000\129\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\224\000\117\000\007\000\138\000\255\255\225\000\
    \131\000\144\000\255\255\140\000\143\000\118\000\151\000\147\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \226\000\008\000\131\000\159\000\132\000\164\000\163\000\255\255\
    \109\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\165\000\181\000\180\000\179\000\008\000\
    \176\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\222\000\174\000\007\000\175\000\177\000\
    \223\000\255\255\178\000\185\000\188\000\203\000\202\000\216\000\
    \212\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\131\000\255\255\132\000\255\255\255\255\
    \255\255\255\255\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\255\255\239\000\062\001\240\000\
    \008\000\052\001\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\220\000\239\000\007\000\240\000\
    \255\255\221\000\051\001\255\255\161\000\255\255\043\001\160\000\
    \255\255\042\001\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\131\000\240\000\132\000\162\000\
    \041\001\111\000\027\001\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\240\000\255\255\255\255\
    \255\255\008\000\255\255\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\158\000\007\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\
    \157\000\255\255\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\131\000\255\255\132\000\255\255\
    \024\001\049\001\030\001\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\255\255\255\255\044\001\
    \045\001\008\000\110\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\155\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\007\000\040\001\065\000\
    \050\001\054\000\255\255\031\001\255\255\255\255\255\255\255\255\
    \046\001\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\131\000\058\001\132\000\065\000\048\000\
    \054\000\064\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\133\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\039\001\038\001\032\001\033\001\
    \008\000\112\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\028\001\042\000\041\000\041\000\
    \041\000\041\000\062\000\059\000\065\000\099\000\108\000\060\000\
    \099\000\048\001\058\000\054\001\069\001\065\001\061\000\075\001\
    \107\000\046\000\071\001\063\000\042\000\087\000\255\255\043\000\
    \255\255\241\000\074\001\065\000\099\000\082\001\098\000\099\000\
    \047\001\098\000\240\000\044\000\255\255\255\255\255\255\078\001\
    \255\255\103\001\096\000\143\001\087\000\096\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\
    \097\000\241\000\025\001\097\000\104\001\029\001\104\001\024\001\
    \118\001\000\000\240\000\166\000\107\001\104\001\000\000\000\000\
    \000\000\103\001\000\000\070\000\168\000\026\001\168\000\255\255\
    \000\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\255\255\104\001\255\255\104\001\062\000\
    \059\000\000\000\070\000\166\000\060\000\104\001\000\000\058\000\
    \000\000\000\000\000\000\061\000\000\000\000\000\046\000\000\000\
    \063\000\000\000\000\000\000\000\113\000\117\001\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\000\000\000\000\000\000\
    \255\255\071\000\255\255\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \255\255\000\000\000\000\000\000\071\000\000\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \077\000\000\000\000\000\000\000\255\255\000\000\255\255\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\255\255\000\000\000\000\000\000\000\000\000\000\077\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\232\000\232\000\232\000\232\000\232\000\232\000\
    \232\000\232\000\232\000\232\000\236\000\236\000\236\000\236\000\
    \236\000\236\000\236\000\236\000\236\000\236\000\000\000\166\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\000\000\000\000\000\000\000\000\078\000\166\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\079\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\000\000\000\000\000\000\
    \000\000\078\000\000\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\085\000\237\000\000\000\
    \237\000\000\000\000\000\236\000\236\000\236\000\236\000\236\000\
    \236\000\236\000\236\000\236\000\236\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\085\000\000\000\000\000\000\000\
    \232\000\000\000\234\000\234\000\234\000\234\000\234\000\234\000\
    \234\000\234\000\234\000\234\000\236\000\236\000\236\000\236\000\
    \236\000\236\000\236\000\236\000\236\000\236\000\000\000\233\000\
    \000\000\000\000\000\000\000\000\000\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\233\000\
    \000\000\000\000\000\000\086\000\000\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\087\000\
    \034\001\034\001\034\001\034\001\034\001\034\001\034\001\034\001\
    \034\001\034\001\022\001\021\001\021\001\021\001\021\001\000\000\
    \000\000\000\000\037\001\000\000\037\001\000\000\087\000\036\001\
    \036\001\036\001\036\001\036\001\036\001\036\001\036\001\036\001\
    \036\001\022\001\000\000\000\000\000\000\000\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \023\001\000\000\000\000\000\000\000\000\000\000\000\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\000\000\000\000\000\000\136\000\086\000\000\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\000\000\000\000\000\000\000\000\136\000\
    \000\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\007\000\000\000\000\000\136\000\111\001\
    \111\001\111\001\111\001\111\001\111\001\111\001\111\001\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\131\000\000\000\132\000\000\000\000\000\000\000\000\000\
    \008\000\008\000\008\000\008\000\134\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\000\000\000\000\000\000\000\000\008\000\000\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\007\000\000\000\034\001\034\001\034\001\034\001\
    \034\001\034\001\034\001\034\001\034\001\034\001\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \131\000\035\001\132\000\000\000\000\000\000\000\000\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\135\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\035\001\000\000\000\000\000\000\008\000\000\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\007\000\036\001\036\001\036\001\036\001\036\001\036\001\
    \036\001\036\001\036\001\036\001\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\131\000\
    \000\000\132\000\000\000\000\000\000\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \000\000\000\000\000\000\000\000\008\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \136\000\036\001\036\001\036\001\036\001\036\001\036\001\036\001\
    \036\001\036\001\036\001\000\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\000\000\000\000\
    \000\000\000\000\137\000\000\000\000\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\000\000\
    \000\000\000\000\000\000\136\000\000\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\139\000\
    \000\000\000\000\136\000\096\001\096\001\096\001\096\001\096\001\
    \096\001\096\001\096\001\096\001\096\001\100\001\100\001\100\001\
    \100\001\100\001\100\001\100\001\100\001\100\001\100\001\000\000\
    \035\001\000\000\000\000\000\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\000\000\000\000\
    \035\001\000\000\139\000\000\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\000\000\
    \000\000\140\000\112\001\112\001\112\001\112\001\112\001\112\001\
    \112\001\112\001\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\147\000\000\000\148\000\000\000\
    \000\000\000\000\000\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\000\000\000\000\000\000\
    \141\000\139\000\000\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\000\000\
    \000\000\000\000\000\000\141\000\000\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \100\001\100\001\100\001\100\001\100\001\100\001\100\001\100\001\
    \100\001\100\001\000\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\143\000\000\000\142\000\
    \000\000\000\000\000\000\000\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\000\000\000\000\
    \000\000\145\000\141\000\000\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \000\000\000\000\000\000\000\000\145\000\000\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\000\000\000\000\145\000\122\001\122\001\122\001\122\001\
    \122\001\122\001\122\001\122\001\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\000\000\000\000\
    \000\000\000\000\146\000\000\000\000\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\000\000\
    \000\000\000\000\000\000\145\000\000\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\149\000\
    \000\000\101\001\145\000\101\001\000\000\000\000\100\001\100\001\
    \100\001\100\001\100\001\100\001\100\001\100\001\100\001\100\001\
    \000\000\000\000\000\000\000\000\110\001\110\001\110\001\110\001\
    \110\001\110\001\110\001\110\001\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\000\000\000\000\
    \000\000\000\000\149\000\000\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\000\000\
    \000\000\149\000\000\000\000\000\109\001\000\000\000\000\000\000\
    \000\000\000\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\000\000\000\000\000\000\000\000\
    \150\000\000\000\000\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\000\000\000\000\000\000\
    \000\000\149\000\000\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\152\000\000\000\096\001\
    \149\000\098\001\098\001\098\001\098\001\098\001\098\001\098\001\
    \098\001\098\001\098\001\123\001\123\001\123\001\123\001\123\001\
    \123\001\123\001\123\001\000\000\000\000\000\000\097\001\000\000\
    \000\000\000\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\255\255\000\000\097\001\000\000\
    \152\000\000\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\134\001\134\001\134\001\
    \134\001\134\001\134\001\134\001\134\001\000\000\000\000\000\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\147\000\000\000\153\000\000\000\000\000\000\000\
    \000\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\000\000\000\000\000\000\154\000\152\000\
    \000\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\000\000\000\000\000\000\
    \000\000\154\000\000\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\000\000\000\000\
    \154\000\135\001\135\001\135\001\135\001\135\001\135\001\135\001\
    \135\001\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\000\000\000\000\000\000\000\000\150\000\
    \000\000\000\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\000\000\000\000\000\000\000\000\
    \154\000\000\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\007\000\000\000\000\000\154\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\131\000\000\000\132\000\000\000\000\000\000\000\
    \000\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\000\000\000\000\000\000\000\000\008\000\
    \000\000\008\000\008\000\008\000\008\000\156\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\007\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\131\000\000\000\132\000\000\000\000\000\000\000\000\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\000\000\000\000\000\000\211\000\008\000\000\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\135\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\000\000\000\000\000\000\000\000\
    \211\000\000\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\212\000\000\000\213\000\000\000\000\000\000\000\
    \000\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\000\000\000\000\000\000\214\000\211\000\
    \000\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\000\000\000\000\000\000\
    \000\000\214\000\000\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\000\000\000\000\
    \214\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\000\000\000\000\000\000\000\000\215\000\
    \000\000\000\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\000\000\000\000\000\000\000\000\
    \214\000\000\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\217\000\000\000\000\000\214\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\000\000\000\000\000\000\000\000\217\000\
    \000\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\212\000\000\000\218\000\000\000\000\000\000\000\000\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\000\000\000\000\000\000\219\000\217\000\000\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\000\000\000\000\000\000\000\000\
    \219\000\000\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\000\000\000\000\219\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\000\000\000\000\000\000\000\000\215\000\000\000\
    \000\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\000\000\000\000\000\000\000\000\219\000\
    \000\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\227\000\000\000\007\000\219\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\131\000\000\000\132\000\000\000\000\000\
    \000\000\000\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\000\000\000\000\000\000\000\000\
    \008\000\000\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\238\000\238\000\238\000\238\000\
    \238\000\238\000\238\000\238\000\238\000\238\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\238\000\238\000\238\000\
    \238\000\238\000\238\000\000\000\000\000\232\000\000\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\234\000\
    \234\000\000\000\129\001\129\001\129\001\129\001\141\001\141\001\
    \141\001\141\001\000\000\000\000\233\000\238\000\238\000\238\000\
    \238\000\238\000\238\000\229\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\130\001\000\000\230\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\142\001\140\001\
    \000\000\000\000\000\000\000\000\233\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\229\000\121\001\121\001\121\001\121\001\
    \121\001\121\001\121\001\121\001\230\000\238\000\238\000\238\000\
    \238\000\238\000\238\000\238\000\238\000\238\000\238\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\238\000\238\000\
    \238\000\238\000\238\000\238\000\128\001\000\000\000\000\000\000\
    \000\000\229\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\230\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\238\000\238\000\
    \238\000\238\000\238\000\238\000\022\001\021\001\021\001\021\001\
    \021\001\229\000\000\000\000\000\120\001\000\000\000\000\000\000\
    \000\000\000\000\230\000\133\001\133\001\133\001\133\001\133\001\
    \133\001\133\001\133\001\022\001\017\001\246\000\253\000\248\000\
    \003\001\001\001\247\000\016\001\015\001\005\001\004\001\008\001\
    \010\001\009\001\020\001\245\000\244\000\244\000\244\000\244\000\
    \244\000\244\000\244\000\244\000\244\000\254\000\007\001\006\001\
    \018\001\002\001\000\000\019\001\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\252\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \251\000\249\000\249\000\249\000\249\000\249\000\014\001\000\000\
    \013\001\255\000\249\000\132\001\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \250\000\249\000\249\000\249\000\249\000\249\000\012\001\000\001\
    \011\001\096\001\000\000\244\000\244\000\244\000\244\000\244\000\
    \244\000\244\000\244\000\244\000\244\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\255\255\000\000\000\000\092\001\
    \097\001\000\000\000\000\000\000\255\255\000\000\000\000\093\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\096\001\
    \094\001\099\001\099\001\099\001\099\001\099\001\099\001\099\001\
    \099\001\098\001\098\001\000\000\000\000\000\000\000\000\092\001\
    \097\001\000\000\000\000\000\000\000\000\092\001\097\001\093\001\
    \000\000\000\000\000\000\000\000\000\000\093\001\000\000\000\000\
    \094\001\000\000\000\000\000\000\000\000\000\000\094\001\000\000\
    \000\000\095\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\092\001\097\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\093\001\000\000\248\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\094\001\000\000\
    \000\000\095\001\000\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\065\001\000\000\066\001\
    \000\000\000\000\000\000\255\255\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\000\000\000\000\
    \000\000\000\000\248\000\000\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\065\001\000\000\066\001\000\000\
    \000\000\000\000\000\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\000\000\000\000\000\000\
    \000\000\249\000\000\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\088\001\000\000\248\000\
    \000\000\000\000\089\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\090\001\249\000\065\001\000\000\066\001\
    \000\000\000\000\000\000\000\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\000\000\000\000\
    \000\000\000\000\249\000\000\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\086\001\000\000\
    \248\000\000\000\000\000\087\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\065\001\000\000\
    \066\001\000\000\000\000\000\000\000\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\000\000\
    \000\000\000\000\000\000\249\000\000\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\063\001\
    \000\000\248\000\000\000\000\000\064\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\065\001\
    \000\000\066\001\000\000\000\000\000\000\000\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \000\000\000\000\000\000\053\001\249\000\000\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\000\000\000\000\000\000\000\000\053\001\000\000\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \054\001\000\000\055\001\000\000\000\000\000\000\000\000\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\000\000\000\000\000\000\056\001\053\001\000\000\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\000\000\000\000\000\000\000\000\056\001\
    \000\000\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\000\000\000\000\056\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\000\000\000\000\000\000\000\000\057\001\000\000\000\000\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\000\000\000\000\000\000\000\000\056\001\000\000\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\059\001\000\000\000\000\056\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\000\000\000\000\000\000\000\000\059\001\000\000\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\054\001\
    \000\000\060\001\000\000\000\000\000\000\000\000\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \000\000\000\000\000\000\061\001\059\001\000\000\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\000\000\000\000\000\000\000\000\061\001\000\000\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\000\000\000\000\061\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \000\000\000\000\000\000\000\000\057\001\000\000\000\000\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\000\000\000\000\000\000\000\000\061\001\000\000\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\067\001\000\000\000\000\061\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \000\000\000\000\000\000\000\000\067\001\000\000\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\000\000\000\000\067\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\000\000\000\000\
    \000\000\000\000\068\001\000\000\000\000\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\000\000\
    \000\000\000\000\000\000\067\001\000\000\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\070\001\
    \000\000\000\000\067\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\000\000\000\000\
    \000\000\000\000\070\001\000\000\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\000\000\
    \000\000\071\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\078\001\000\000\079\001\000\000\
    \000\000\000\000\000\000\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\000\000\000\000\000\000\
    \072\001\070\001\000\000\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\000\000\
    \000\000\000\000\000\000\072\001\000\000\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\074\001\000\000\073\001\
    \000\000\000\000\000\000\000\000\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\000\000\000\000\
    \000\000\076\001\072\001\000\000\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \000\000\000\000\000\000\000\000\076\001\000\000\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\000\000\000\000\076\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\000\000\000\000\
    \000\000\000\000\077\001\000\000\000\000\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\000\000\
    \000\000\000\000\000\000\076\001\000\000\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\080\001\
    \000\000\000\000\076\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\000\000\000\000\
    \000\000\000\000\080\001\000\000\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\000\000\
    \000\000\080\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\000\000\000\000\000\000\000\000\
    \081\001\000\000\000\000\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\000\000\000\000\000\000\
    \000\000\080\001\000\000\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\083\001\000\000\000\000\
    \080\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\000\000\000\000\000\000\000\000\
    \083\001\000\000\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\078\001\000\000\084\001\000\000\000\000\000\000\
    \000\000\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\000\000\000\000\000\000\085\001\083\001\
    \000\000\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\000\000\000\000\000\000\
    \000\000\085\001\000\000\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\000\000\000\000\
    \085\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\000\000\000\000\000\000\000\000\081\001\
    \000\000\000\000\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\000\000\000\000\000\000\000\000\
    \085\001\000\000\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\091\001\000\000\248\000\085\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\065\001\000\000\066\001\000\000\
    \000\000\000\000\000\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\000\000\000\000\000\000\
    \000\000\249\000\000\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\102\001\102\001\102\001\
    \102\001\102\001\102\001\102\001\102\001\102\001\102\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\102\001\102\001\
    \102\001\102\001\102\001\102\001\000\000\000\000\096\001\000\000\
    \099\001\099\001\099\001\099\001\099\001\099\001\099\001\099\001\
    \098\001\098\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\097\001\102\001\102\001\
    \102\001\102\001\102\001\102\001\093\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\094\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\097\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\093\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\094\001\102\001\102\001\
    \102\001\102\001\102\001\102\001\102\001\102\001\102\001\102\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\102\001\
    \102\001\102\001\102\001\102\001\102\001\000\000\000\000\000\000\
    \000\000\000\000\093\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\094\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\102\001\
    \102\001\102\001\102\001\102\001\102\001\000\000\000\000\000\000\
    \000\000\000\000\093\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\094\001\113\001\113\001\113\001\113\001\
    \113\001\113\001\113\001\113\001\113\001\113\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\113\001\113\001\113\001\
    \113\001\113\001\113\001\114\001\114\001\114\001\114\001\114\001\
    \114\001\114\001\114\001\114\001\114\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\114\001\114\001\114\001\114\001\
    \114\001\114\001\000\000\000\000\000\000\113\001\113\001\113\001\
    \113\001\113\001\113\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\124\001\124\001\124\001\124\001\124\001\124\001\124\001\
    \124\001\124\001\124\001\000\000\114\001\114\001\114\001\114\001\
    \114\001\114\001\124\001\124\001\124\001\124\001\124\001\124\001\
    \125\001\125\001\125\001\125\001\125\001\125\001\125\001\125\001\
    \125\001\125\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\125\001\125\001\125\001\125\001\125\001\125\001\000\000\
    \000\000\000\000\124\001\124\001\124\001\124\001\124\001\124\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\136\001\136\001\
    \136\001\136\001\136\001\136\001\136\001\136\001\136\001\136\001\
    \000\000\125\001\125\001\125\001\125\001\125\001\125\001\136\001\
    \136\001\136\001\136\001\136\001\136\001\137\001\137\001\137\001\
    \137\001\137\001\137\001\137\001\137\001\137\001\137\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\137\001\137\001\
    \137\001\137\001\137\001\137\001\000\000\000\000\000\000\136\001\
    \136\001\136\001\136\001\136\001\136\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\137\001\137\001\
    \137\001\137\001\137\001\137\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\190\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \012\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \013\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\014\000\015\000\015\000\017\000\018\000\
    \018\000\018\000\019\000\027\000\030\000\003\000\003\000\038\000\
    \044\000\046\000\049\000\050\000\051\000\003\000\052\000\032\000\
    \045\000\053\000\014\000\033\000\048\000\004\000\003\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\032\000\058\000\033\000\033\000\003\000\003\000\034\000\
    \053\000\055\000\034\000\004\000\004\000\003\000\061\000\060\000\
    \040\000\062\000\063\000\004\000\066\000\040\000\003\000\060\000\
    \034\000\034\000\028\000\034\000\004\000\028\000\067\000\004\000\
    \068\000\072\000\027\000\040\000\047\000\073\000\043\000\059\000\
    \074\000\059\000\075\000\004\000\004\000\059\000\080\000\081\000\
    \082\000\083\000\090\000\004\000\007\000\091\000\028\000\069\000\
    \028\000\028\000\092\000\047\000\004\000\043\000\043\000\004\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\088\000\007\000\093\000\069\000\094\000\
    \000\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\096\000\104\000\114\000\115\000\007\000\
    \117\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\008\000\028\000\076\000\047\000\084\000\
    \043\000\118\000\097\000\121\000\102\000\122\000\098\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\106\000\008\000\076\000\108\000\084\000\110\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\100\000\124\000\125\000\127\000\008\000\128\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\009\000\116\000\009\000\131\000\045\000\009\000\
    \137\000\143\000\048\000\144\000\146\000\116\000\147\000\150\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\158\000\009\000\160\000\161\000\055\000\
    \108\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\164\000\169\000\170\000\171\000\009\000\
    \172\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\010\000\173\000\010\000\174\000\176\000\
    \010\000\088\000\177\000\183\000\187\000\200\000\201\000\212\000\
    \215\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\097\000\010\000\102\000\107\000\
    \113\000\119\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\098\000\230\000\253\000\239\000\
    \010\000\254\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\011\000\230\000\011\000\239\000\
    \106\000\011\000\255\000\108\000\159\000\110\000\003\001\159\000\
    \100\000\004\001\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\241\000\011\000\159\000\
    \005\001\107\000\019\001\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\241\000\109\000\120\000\
    \123\000\011\000\126\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\035\000\036\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\129\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\185\000\036\000\188\000\
    \023\001\000\001\017\001\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\024\001\111\000\002\001\
    \002\001\036\000\109\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\037\000\039\001\039\000\
    \000\001\054\000\054\000\017\001\089\000\107\000\113\000\119\000\
    \045\001\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\054\001\037\000\039\000\039\000\
    \054\000\039\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\006\001\006\001\010\001\010\001\
    \037\000\111\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\018\001\042\000\042\000\042\000\
    \042\000\042\000\039\000\039\000\065\000\095\000\089\000\039\000\
    \099\000\001\001\039\000\057\001\065\001\068\001\039\000\074\001\
    \089\000\039\000\075\001\039\000\042\000\087\000\087\000\042\000\
    \112\000\229\000\077\001\065\000\095\000\078\001\095\000\099\000\
    \001\001\099\000\229\000\042\000\109\000\120\000\123\000\081\001\
    \126\000\094\001\095\000\142\001\087\000\099\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\
    \095\000\229\000\020\001\099\000\093\001\018\001\103\001\020\001\
    \115\001\255\255\229\000\157\000\105\001\093\001\255\255\255\255\
    \255\255\094\001\255\255\070\000\166\000\020\001\166\000\129\000\
    \255\255\166\000\166\000\166\000\166\000\166\000\166\000\166\000\
    \166\000\166\000\166\000\185\000\093\001\188\000\103\001\065\000\
    \065\000\255\255\070\000\157\000\065\000\093\001\255\255\065\000\
    \255\255\255\255\255\255\065\000\255\255\255\255\065\000\255\255\
    \065\000\255\255\255\255\255\255\112\000\115\001\255\255\255\255\
    \255\255\255\255\255\255\024\001\111\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\255\255\255\255\255\255\
    \054\000\070\000\089\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \105\001\255\255\255\255\255\255\071\000\255\255\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \077\000\255\255\255\255\255\255\087\000\255\255\112\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\115\001\255\255\255\255\255\255\255\255\255\255\077\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\232\000\232\000\232\000\232\000\232\000\232\000\
    \232\000\232\000\232\000\232\000\236\000\236\000\236\000\236\000\
    \236\000\236\000\236\000\236\000\236\000\236\000\255\255\232\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\255\255\255\255\255\255\255\255\077\000\232\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\078\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\255\255\255\255\255\255\
    \255\255\078\000\255\255\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\085\000\233\000\255\255\
    \233\000\255\255\255\255\233\000\233\000\233\000\233\000\233\000\
    \233\000\233\000\233\000\233\000\233\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\085\000\255\255\255\255\255\255\
    \234\000\255\255\234\000\234\000\234\000\234\000\234\000\234\000\
    \234\000\234\000\234\000\234\000\237\000\237\000\237\000\237\000\
    \237\000\237\000\237\000\237\000\237\000\237\000\255\255\234\000\
    \255\255\255\255\255\255\255\255\255\255\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\234\000\
    \255\255\255\255\255\255\085\000\255\255\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\086\000\
    \009\001\009\001\009\001\009\001\009\001\009\001\009\001\009\001\
    \009\001\009\001\022\001\022\001\022\001\022\001\022\001\255\255\
    \255\255\255\255\035\001\255\255\035\001\255\255\086\000\035\001\
    \035\001\035\001\035\001\035\001\035\001\035\001\035\001\035\001\
    \035\001\022\001\255\255\255\255\255\255\255\255\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \022\001\255\255\255\255\255\255\255\255\255\255\255\255\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\255\255\255\255\255\255\132\000\086\000\255\255\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\255\255\255\255\255\255\255\255\132\000\
    \255\255\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\133\000\255\255\255\255\132\000\110\001\
    \110\001\110\001\110\001\110\001\110\001\110\001\110\001\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\255\255\133\000\255\255\255\255\255\255\255\255\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\255\255\255\255\255\255\255\255\133\000\255\255\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\134\000\255\255\034\001\034\001\034\001\034\001\
    \034\001\034\001\034\001\034\001\034\001\034\001\134\000\134\000\
    \134\000\134\000\134\000\134\000\134\000\134\000\134\000\134\000\
    \134\000\034\001\134\000\255\255\255\255\255\255\255\255\134\000\
    \134\000\134\000\134\000\134\000\134\000\134\000\134\000\134\000\
    \134\000\134\000\134\000\134\000\134\000\134\000\134\000\134\000\
    \134\000\134\000\134\000\134\000\134\000\134\000\134\000\134\000\
    \134\000\034\001\255\255\255\255\255\255\134\000\255\255\134\000\
    \134\000\134\000\134\000\134\000\134\000\134\000\134\000\134\000\
    \134\000\134\000\134\000\134\000\134\000\134\000\134\000\134\000\
    \134\000\134\000\134\000\134\000\134\000\134\000\134\000\134\000\
    \134\000\135\000\036\001\036\001\036\001\036\001\036\001\036\001\
    \036\001\036\001\036\001\036\001\255\255\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \255\255\135\000\255\255\255\255\255\255\255\255\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \255\255\255\255\255\255\255\255\135\000\255\255\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \136\000\037\001\037\001\037\001\037\001\037\001\037\001\037\001\
    \037\001\037\001\037\001\255\255\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\255\255\255\255\
    \255\255\255\255\136\000\255\255\255\255\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\255\255\
    \255\255\255\255\255\255\136\000\255\255\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\138\000\
    \255\255\255\255\136\000\096\001\096\001\096\001\096\001\096\001\
    \096\001\096\001\096\001\096\001\096\001\100\001\100\001\100\001\
    \100\001\100\001\100\001\100\001\100\001\100\001\100\001\255\255\
    \096\001\255\255\255\255\255\255\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\255\255\255\255\
    \096\001\255\255\138\000\255\255\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\139\000\255\255\
    \255\255\138\000\111\001\111\001\111\001\111\001\111\001\111\001\
    \111\001\111\001\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\255\255\139\000\255\255\
    \255\255\255\255\255\255\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\255\255\255\255\255\255\
    \140\000\139\000\255\255\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\140\000\140\000\140\000\
    \140\000\140\000\140\000\140\000\140\000\140\000\140\000\140\000\
    \140\000\140\000\140\000\140\000\140\000\140\000\140\000\140\000\
    \140\000\140\000\140\000\140\000\140\000\140\000\140\000\255\255\
    \255\255\255\255\255\255\140\000\255\255\140\000\140\000\140\000\
    \140\000\140\000\140\000\140\000\140\000\140\000\140\000\140\000\
    \140\000\140\000\140\000\140\000\140\000\140\000\140\000\140\000\
    \140\000\140\000\140\000\140\000\140\000\140\000\140\000\141\000\
    \101\001\101\001\101\001\101\001\101\001\101\001\101\001\101\001\
    \101\001\101\001\255\255\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\255\255\141\000\
    \255\255\255\255\255\255\255\255\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\255\255\255\255\
    \255\255\142\000\141\000\255\255\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\
    \141\000\141\000\141\000\141\000\141\000\141\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \255\255\255\255\255\255\255\255\142\000\255\255\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \145\000\255\255\255\255\142\000\121\001\121\001\121\001\121\001\
    \121\001\121\001\121\001\121\001\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\255\255\255\255\
    \255\255\255\255\145\000\255\255\255\255\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\255\255\
    \255\255\255\255\255\255\145\000\255\255\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\148\000\
    \255\255\097\001\145\000\097\001\255\255\255\255\097\001\097\001\
    \097\001\097\001\097\001\097\001\097\001\097\001\097\001\097\001\
    \255\255\255\255\255\255\255\255\107\001\107\001\107\001\107\001\
    \107\001\107\001\107\001\107\001\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\255\255\255\255\
    \255\255\255\255\148\000\255\255\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\149\000\255\255\
    \255\255\148\000\255\255\255\255\107\001\255\255\255\255\255\255\
    \255\255\255\255\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\255\255\255\255\255\255\255\255\
    \149\000\255\255\255\255\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\255\255\255\255\255\255\
    \255\255\149\000\255\255\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\151\000\255\255\098\001\
    \149\000\098\001\098\001\098\001\098\001\098\001\098\001\098\001\
    \098\001\098\001\098\001\122\001\122\001\122\001\122\001\122\001\
    \122\001\122\001\122\001\255\255\255\255\255\255\098\001\255\255\
    \255\255\255\255\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\107\001\255\255\098\001\255\255\
    \151\000\255\255\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\152\000\133\001\133\001\133\001\
    \133\001\133\001\133\001\133\001\133\001\255\255\255\255\255\255\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\255\255\152\000\255\255\255\255\255\255\
    \255\255\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\255\255\255\255\255\255\153\000\152\000\
    \255\255\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\255\255\255\255\255\255\
    \255\255\153\000\255\255\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\154\000\255\255\255\255\
    \153\000\134\001\134\001\134\001\134\001\134\001\134\001\134\001\
    \134\001\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\255\255\255\255\255\255\255\255\154\000\
    \255\255\255\255\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\255\255\255\255\255\255\255\255\
    \154\000\255\255\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\155\000\255\255\255\255\154\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\255\255\155\000\255\255\255\255\255\255\
    \255\255\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\255\255\255\255\255\255\255\255\155\000\
    \255\255\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\156\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\255\255\156\000\255\255\255\255\255\255\255\255\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\255\255\255\255\255\255\210\000\156\000\255\255\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\210\000\210\000\210\000\210\000\210\000\210\000\
    \210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\
    \210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\
    \210\000\210\000\210\000\210\000\255\255\255\255\255\255\255\255\
    \210\000\255\255\210\000\210\000\210\000\210\000\210\000\210\000\
    \210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\
    \210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\
    \210\000\210\000\210\000\210\000\211\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\255\255\211\000\255\255\255\255\255\255\
    \255\255\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\255\255\255\255\255\255\213\000\211\000\
    \255\255\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\213\000\213\000\213\000\213\000\213\000\
    \213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\
    \213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\
    \213\000\213\000\213\000\213\000\213\000\255\255\255\255\255\255\
    \255\255\213\000\255\255\213\000\213\000\213\000\213\000\213\000\
    \213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\
    \213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\
    \213\000\213\000\213\000\213\000\213\000\214\000\255\255\255\255\
    \213\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\255\255\255\255\255\255\255\255\214\000\
    \255\255\255\255\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\255\255\255\255\255\255\255\255\
    \214\000\255\255\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\216\000\255\255\255\255\214\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\216\000\216\000\216\000\216\000\216\000\216\000\216\000\
    \216\000\216\000\216\000\216\000\216\000\216\000\216\000\216\000\
    \216\000\216\000\216\000\216\000\216\000\216\000\216\000\216\000\
    \216\000\216\000\216\000\255\255\255\255\255\255\255\255\216\000\
    \255\255\216\000\216\000\216\000\216\000\216\000\216\000\216\000\
    \216\000\216\000\216\000\216\000\216\000\216\000\216\000\216\000\
    \216\000\216\000\216\000\216\000\216\000\216\000\216\000\216\000\
    \216\000\216\000\216\000\217\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\255\255\217\000\255\255\255\255\255\255\255\255\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\255\255\255\255\255\255\218\000\217\000\255\255\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\255\255\255\255\255\255\255\255\
    \218\000\255\255\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\219\000\255\255\255\255\218\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\255\255\255\255\255\255\255\255\219\000\255\255\
    \255\255\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\255\255\255\255\255\255\255\255\219\000\
    \255\255\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\226\000\255\255\226\000\219\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\226\000\226\000\226\000\226\000\226\000\226\000\226\000\
    \226\000\226\000\226\000\226\000\255\255\226\000\255\255\255\255\
    \255\255\255\255\226\000\226\000\226\000\226\000\226\000\226\000\
    \226\000\226\000\226\000\226\000\226\000\226\000\226\000\226\000\
    \226\000\226\000\226\000\226\000\226\000\226\000\226\000\226\000\
    \226\000\226\000\226\000\226\000\255\255\255\255\255\255\255\255\
    \226\000\255\255\226\000\226\000\226\000\226\000\226\000\226\000\
    \226\000\226\000\226\000\226\000\226\000\226\000\226\000\226\000\
    \226\000\226\000\226\000\226\000\226\000\226\000\226\000\226\000\
    \226\000\226\000\226\000\226\000\231\000\231\000\231\000\231\000\
    \231\000\231\000\231\000\231\000\231\000\231\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\231\000\231\000\231\000\
    \231\000\231\000\231\000\255\255\255\255\235\000\255\255\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\255\255\126\001\126\001\126\001\126\001\138\001\138\001\
    \138\001\138\001\255\255\255\255\235\000\231\000\231\000\231\000\
    \231\000\231\000\231\000\235\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\126\001\255\255\235\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\138\001\138\001\
    \255\255\255\255\255\255\255\255\235\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\235\000\117\001\117\001\117\001\117\001\
    \117\001\117\001\117\001\117\001\235\000\238\000\238\000\238\000\
    \238\000\238\000\238\000\238\000\238\000\238\000\238\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\238\000\238\000\
    \238\000\238\000\238\000\238\000\126\001\255\255\255\255\255\255\
    \255\255\238\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\238\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\238\000\238\000\
    \238\000\238\000\238\000\238\000\242\000\242\000\242\000\242\000\
    \242\000\238\000\255\255\255\255\117\001\255\255\255\255\255\255\
    \255\255\255\255\238\000\128\001\128\001\128\001\128\001\128\001\
    \128\001\128\001\128\001\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\255\255\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\255\255\
    \242\000\242\000\242\000\128\001\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\244\000\255\255\244\000\244\000\244\000\244\000\244\000\
    \244\000\244\000\244\000\244\000\244\000\255\255\255\255\255\255\
    \126\001\255\255\255\255\255\255\138\001\255\255\255\255\244\000\
    \244\000\255\255\255\255\255\255\117\001\255\255\255\255\244\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\245\000\
    \244\000\245\000\245\000\245\000\245\000\245\000\245\000\245\000\
    \245\000\245\000\245\000\255\255\255\255\255\255\255\255\244\000\
    \244\000\255\255\255\255\255\255\255\255\245\000\245\000\244\000\
    \255\255\255\255\255\255\255\255\255\255\245\000\255\255\255\255\
    \244\000\255\255\255\255\255\255\255\255\255\255\245\000\255\255\
    \255\255\245\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\245\000\245\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\245\000\255\255\248\000\
    \255\255\255\255\255\255\128\001\255\255\255\255\245\000\255\255\
    \255\255\245\000\255\255\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\255\255\248\000\
    \255\255\255\255\255\255\242\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\255\255\255\255\
    \255\255\255\255\248\000\255\255\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\
    \248\000\248\000\248\000\248\000\248\000\248\000\249\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\255\255\249\000\255\255\
    \255\255\255\255\255\255\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\255\255\255\255\255\255\
    \255\255\249\000\255\255\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\249\000\249\000\249\000\250\000\255\255\250\000\
    \255\255\255\255\250\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\250\000\250\000\250\000\250\000\250\000\
    \250\000\250\000\250\000\250\000\250\000\250\000\255\255\250\000\
    \255\255\255\255\255\255\255\255\250\000\250\000\250\000\250\000\
    \250\000\250\000\250\000\250\000\250\000\250\000\250\000\250\000\
    \250\000\250\000\250\000\250\000\250\000\250\000\250\000\250\000\
    \250\000\250\000\250\000\250\000\250\000\250\000\255\255\255\255\
    \255\255\255\255\250\000\255\255\250\000\250\000\250\000\250\000\
    \250\000\250\000\250\000\250\000\250\000\250\000\250\000\250\000\
    \250\000\250\000\250\000\250\000\250\000\250\000\250\000\250\000\
    \250\000\250\000\250\000\250\000\250\000\250\000\251\000\255\255\
    \251\000\255\255\255\255\251\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\251\000\251\000\251\000\251\000\
    \251\000\251\000\251\000\251\000\251\000\251\000\251\000\255\255\
    \251\000\255\255\255\255\255\255\255\255\251\000\251\000\251\000\
    \251\000\251\000\251\000\251\000\251\000\251\000\251\000\251\000\
    \251\000\251\000\251\000\251\000\251\000\251\000\251\000\251\000\
    \251\000\251\000\251\000\251\000\251\000\251\000\251\000\255\255\
    \255\255\255\255\255\255\251\000\255\255\251\000\251\000\251\000\
    \251\000\251\000\251\000\251\000\251\000\251\000\251\000\251\000\
    \251\000\251\000\251\000\251\000\251\000\251\000\251\000\251\000\
    \251\000\251\000\251\000\251\000\251\000\251\000\251\000\252\000\
    \255\255\252\000\255\255\255\255\252\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\252\000\252\000\252\000\
    \252\000\252\000\252\000\252\000\252\000\252\000\252\000\252\000\
    \255\255\252\000\255\255\255\255\255\255\255\255\252\000\252\000\
    \252\000\252\000\252\000\252\000\252\000\252\000\252\000\252\000\
    \252\000\252\000\252\000\252\000\252\000\252\000\252\000\252\000\
    \252\000\252\000\252\000\252\000\252\000\252\000\252\000\252\000\
    \255\255\255\255\255\255\052\001\252\000\255\255\252\000\252\000\
    \252\000\252\000\252\000\252\000\252\000\252\000\252\000\252\000\
    \252\000\252\000\252\000\252\000\252\000\252\000\252\000\252\000\
    \252\000\252\000\252\000\252\000\252\000\252\000\252\000\252\000\
    \052\001\052\001\052\001\052\001\052\001\052\001\052\001\052\001\
    \052\001\052\001\052\001\052\001\052\001\052\001\052\001\052\001\
    \052\001\052\001\052\001\052\001\052\001\052\001\052\001\052\001\
    \052\001\052\001\255\255\255\255\255\255\255\255\052\001\255\255\
    \052\001\052\001\052\001\052\001\052\001\052\001\052\001\052\001\
    \052\001\052\001\052\001\052\001\052\001\052\001\052\001\052\001\
    \052\001\052\001\052\001\052\001\052\001\052\001\052\001\052\001\
    \052\001\052\001\053\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\255\255\053\001\255\255\255\255\255\255\255\255\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\255\255\255\255\255\255\055\001\053\001\255\255\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\053\001\053\001\053\001\053\001\053\001\053\001\053\001\
    \053\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\255\255\255\255\255\255\255\255\055\001\
    \255\255\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\056\001\255\255\255\255\055\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\255\255\255\255\255\255\255\255\056\001\255\255\255\255\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\255\255\255\255\255\255\255\255\056\001\255\255\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\056\001\056\001\056\001\056\001\056\001\056\001\
    \056\001\056\001\058\001\255\255\255\255\056\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\255\255\255\255\255\255\255\255\058\001\255\255\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\059\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \255\255\059\001\255\255\255\255\255\255\255\255\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \255\255\255\255\255\255\060\001\059\001\255\255\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\255\255\255\255\255\255\255\255\060\001\255\255\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\061\001\255\255\255\255\060\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \255\255\255\255\255\255\255\255\061\001\255\255\255\255\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\255\255\255\255\255\255\255\255\061\001\255\255\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\066\001\255\255\255\255\061\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\066\001\066\001\
    \066\001\066\001\066\001\066\001\066\001\066\001\066\001\066\001\
    \066\001\066\001\066\001\066\001\066\001\066\001\066\001\066\001\
    \066\001\066\001\066\001\066\001\066\001\066\001\066\001\066\001\
    \255\255\255\255\255\255\255\255\066\001\255\255\066\001\066\001\
    \066\001\066\001\066\001\066\001\066\001\066\001\066\001\066\001\
    \066\001\066\001\066\001\066\001\066\001\066\001\066\001\066\001\
    \066\001\066\001\066\001\066\001\066\001\066\001\066\001\066\001\
    \067\001\255\255\255\255\066\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\255\255\255\255\
    \255\255\255\255\067\001\255\255\255\255\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\255\255\
    \255\255\255\255\255\255\067\001\255\255\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\069\001\
    \255\255\255\255\067\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\069\001\069\001\069\001\069\001\
    \069\001\069\001\069\001\069\001\069\001\069\001\069\001\069\001\
    \069\001\069\001\069\001\069\001\069\001\069\001\069\001\069\001\
    \069\001\069\001\069\001\069\001\069\001\069\001\255\255\255\255\
    \255\255\255\255\069\001\255\255\069\001\069\001\069\001\069\001\
    \069\001\069\001\069\001\069\001\069\001\069\001\069\001\069\001\
    \069\001\069\001\069\001\069\001\069\001\069\001\069\001\069\001\
    \069\001\069\001\069\001\069\001\069\001\069\001\070\001\255\255\
    \255\255\069\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\255\255\070\001\255\255\
    \255\255\255\255\255\255\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\255\255\255\255\255\255\
    \071\001\070\001\255\255\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\255\255\
    \255\255\255\255\255\255\071\001\255\255\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\072\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\255\255\072\001\
    \255\255\255\255\255\255\255\255\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\255\255\255\255\
    \255\255\073\001\072\001\255\255\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \255\255\255\255\255\255\255\255\073\001\255\255\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \076\001\255\255\255\255\073\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\255\255\255\255\
    \255\255\255\255\076\001\255\255\255\255\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\255\255\
    \255\255\255\255\255\255\076\001\255\255\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\076\001\076\001\076\001\079\001\
    \255\255\255\255\076\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\079\001\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\079\001\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\079\001\079\001\079\001\255\255\255\255\
    \255\255\255\255\079\001\255\255\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\079\001\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\079\001\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\079\001\079\001\079\001\080\001\255\255\
    \255\255\079\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\255\255\255\255\255\255\255\255\
    \080\001\255\255\255\255\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\255\255\255\255\255\255\
    \255\255\080\001\255\255\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\080\001\080\001\082\001\255\255\255\255\
    \080\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\082\001\082\001\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\255\255\255\255\255\255\255\255\
    \082\001\255\255\082\001\082\001\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\083\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\255\255\083\001\255\255\255\255\255\255\
    \255\255\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\255\255\255\255\255\255\084\001\083\001\
    \255\255\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\255\255\255\255\255\255\
    \255\255\084\001\255\255\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\085\001\255\255\255\255\
    \084\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\255\255\255\255\255\255\255\255\085\001\
    \255\255\255\255\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\255\255\255\255\255\255\255\255\
    \085\001\255\255\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\085\001\085\001\090\001\255\255\090\001\085\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\090\001\090\001\090\001\090\001\090\001\090\001\
    \090\001\090\001\090\001\090\001\090\001\255\255\090\001\255\255\
    \255\255\255\255\255\255\090\001\090\001\090\001\090\001\090\001\
    \090\001\090\001\090\001\090\001\090\001\090\001\090\001\090\001\
    \090\001\090\001\090\001\090\001\090\001\090\001\090\001\090\001\
    \090\001\090\001\090\001\090\001\090\001\255\255\255\255\255\255\
    \255\255\090\001\255\255\090\001\090\001\090\001\090\001\090\001\
    \090\001\090\001\090\001\090\001\090\001\090\001\090\001\090\001\
    \090\001\090\001\090\001\090\001\090\001\090\001\090\001\090\001\
    \090\001\090\001\090\001\090\001\090\001\095\001\095\001\095\001\
    \095\001\095\001\095\001\095\001\095\001\095\001\095\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\095\001\095\001\
    \095\001\095\001\095\001\095\001\255\255\255\255\099\001\255\255\
    \099\001\099\001\099\001\099\001\099\001\099\001\099\001\099\001\
    \099\001\099\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\099\001\095\001\095\001\
    \095\001\095\001\095\001\095\001\099\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\099\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\099\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\099\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\099\001\102\001\102\001\
    \102\001\102\001\102\001\102\001\102\001\102\001\102\001\102\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\102\001\
    \102\001\102\001\102\001\102\001\102\001\255\255\255\255\255\255\
    \255\255\255\255\102\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\102\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\102\001\
    \102\001\102\001\102\001\102\001\102\001\255\255\255\255\255\255\
    \255\255\255\255\102\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\102\001\109\001\109\001\109\001\109\001\
    \109\001\109\001\109\001\109\001\109\001\109\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\109\001\109\001\109\001\
    \109\001\109\001\109\001\113\001\113\001\113\001\113\001\113\001\
    \113\001\113\001\113\001\113\001\113\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\113\001\113\001\113\001\113\001\
    \113\001\113\001\255\255\255\255\255\255\109\001\109\001\109\001\
    \109\001\109\001\109\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\120\001\120\001\120\001\120\001\120\001\120\001\120\001\
    \120\001\120\001\120\001\255\255\113\001\113\001\113\001\113\001\
    \113\001\113\001\120\001\120\001\120\001\120\001\120\001\120\001\
    \124\001\124\001\124\001\124\001\124\001\124\001\124\001\124\001\
    \124\001\124\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\124\001\124\001\124\001\124\001\124\001\124\001\255\255\
    \255\255\255\255\120\001\120\001\120\001\120\001\120\001\120\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\132\001\132\001\
    \132\001\132\001\132\001\132\001\132\001\132\001\132\001\132\001\
    \255\255\124\001\124\001\124\001\124\001\124\001\124\001\132\001\
    \132\001\132\001\132\001\132\001\132\001\136\001\136\001\136\001\
    \136\001\136\001\136\001\136\001\136\001\136\001\136\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\136\001\136\001\
    \136\001\136\001\136\001\136\001\255\255\255\255\255\255\132\001\
    \132\001\132\001\132\001\132\001\132\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\136\001\136\001\
    \136\001\136\001\136\001\136\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\002\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\004\000\005\000\000\000\043\000\
    \000\000\000\000\000\000\000\000\006\000\038\000\152\000\007\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\008\000\000\000\000\000\
    \011\000\000\000\000\000\000\000\000\000\014\000\014\000\014\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\023\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\038\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\060\000\
    \060\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\001\000\020\000\020\000\033\000\033\000\051\000\
    \057\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\001\000\020\000\020\000\033\000\033\000\051\000\057\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\051\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\051\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\048\000\000\000\000\000\000\000\028\000\017\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\000\000\000\000\000\000\000\000\054\000\000\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\057\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \057\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\000\000\000\000\000\000\000\000\054\000\
    \000\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\042\000\069\000\070\000\076\000\077\000\084\000\
    \087\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\042\000\069\000\070\000\076\000\077\000\084\000\087\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\085\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\085\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\083\000\255\255\255\255\255\255\075\000\068\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\255\255\255\255\255\255\255\255\085\000\255\255\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\086\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \086\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\255\255\255\255\255\255\255\255\086\000\
    \255\255\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255";
  Lexing.lex_code =
   "\255\006\255\005\255\004\255\255\000\004\255\000\005\255\000\006\
    \255\007\255\255\008\255\255\000\007\001\008\255\010\255\009\255\
    \255\012\255\011\255\255\000\009\001\011\255\000\010\001\012\255\
    \013\255\255\014\255\255\015\255\255\016\255\255\003\016\001\014\
    \000\013\002\015\255";
}

let rec token lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 17 (-1);(* L=3 [6] <- p ; [5] <- p ; [4] <- p ;  *)
  lexbuf.Lexing.lex_mem.(6) <- lexbuf.Lexing.lex_curr_pos ;
  lexbuf.Lexing.lex_mem.(5) <- lexbuf.Lexing.lex_curr_pos ;
  lexbuf.Lexing.lex_mem.(4) <- lexbuf.Lexing.lex_curr_pos ;
 __ocaml_lex_token_rec lexbuf 0
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 639 "parsing_cocci/lexer_cocci.mll"
    ( let cls = !current_line_started in
      if not cls
      then
	begin
	  match !current_line_type with
	    (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
	      (* increment the logical line even though nothing seen *)
	      start_line true;
	      let info = get_current_line_type lexbuf in
	      reset_line lexbuf;
	      TDirective (Ast.Noindent "", info)
	  | _ -> reset_line lexbuf; token lexbuf
	end
      else (reset_line lexbuf; token lexbuf) )
# 2799 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
let
# 654 "parsing_cocci/lexer_cocci.mll"
                      w
# 2805 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 654 "parsing_cocci/lexer_cocci.mll"
                         ( (* collect whitespaces only when inside a rule *)
    start_line false;
    if !Data.in_rule_name || !Data.in_prolog || !Data.in_iso
    then token lexbuf
    else TWhitespace w )
# 2813 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
let
# 660 "parsing_cocci/lexer_cocci.mll"
                                       after
# 2819 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 660 "parsing_cocci/lexer_cocci.mll"
                                              (
    match !current_line_type with
      (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
	let str =
	  if !current_line_started
	  then (tok lexbuf)
	  else after in
	start_line true;
	TDirective (Ast.Indent str, get_current_line_type lexbuf)
    | _ -> start_line false; token lexbuf )
# 2832 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
let
# 671 "parsing_cocci/lexer_cocci.mll"
                                       after
# 2838 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 671 "parsing_cocci/lexer_cocci.mll"
                                              (
     check_context_linetype after;
     start_line false; token lexbuf )
# 2844 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
let
# 675 "parsing_cocci/lexer_cocci.mll"
                                                                                   spatch
# 2850 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 675 "parsing_cocci/lexer_cocci.mll"
                                                                                           (
     check_context_linetype spatch;
     start_line false; token lexbuf )
# 2856 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
# 679 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TArobArob )
# 2861 "parsing_cocci/lexer_cocci.ml"

  | 6 ->
# 680 "parsing_cocci/lexer_cocci.mll"
         ( pass_zero();
	   if !Data.in_rule_name || not !current_line_started
	   then (start_line true; TArob)
	   else (check_minus_context_linetype "@";
		 TPArob (get_current_line_type lexbuf)) )
# 2870 "parsing_cocci/lexer_cocci.ml"

  | 7 ->
# 687 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; check_minus_context_linetype (tok lexbuf);
	TWhen (get_current_line_type lexbuf) )
# 2876 "parsing_cocci/lexer_cocci.ml"

  | 8 ->
# 691 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; check_minus_context_linetype (tok lexbuf);
	TEllipsis (get_current_line_type lexbuf) )
# 2882 "parsing_cocci/lexer_cocci.ml"

  | 9 ->
# 693 "parsing_cocci/lexer_cocci.mll"
           ( start_line true; check_context_linetype (tok lexbuf);
	     TOEllipsis (get_current_line_type lexbuf) )
# 2888 "parsing_cocci/lexer_cocci.ml"

  | 10 ->
# 695 "parsing_cocci/lexer_cocci.mll"
           ( start_line true; check_context_linetype (tok lexbuf);
	     TCEllipsis (get_current_line_type lexbuf) )
# 2894 "parsing_cocci/lexer_cocci.ml"

  | 11 ->
# 697 "parsing_cocci/lexer_cocci.mll"
            ( start_line true; check_minus_context_linetype (tok lexbuf);
	     TPOEllipsis (get_current_line_type lexbuf) )
# 2900 "parsing_cocci/lexer_cocci.ml"

  | 12 ->
# 699 "parsing_cocci/lexer_cocci.mll"
            ( start_line true; check_minus_context_linetype (tok lexbuf);
	     TPCEllipsis (get_current_line_type lexbuf) )
# 2906 "parsing_cocci/lexer_cocci.ml"

  | 13 ->
# 701 "parsing_cocci/lexer_cocci.mll"
        ( pass_zero();
	  if !current_line_started
	  then (start_line true; TMinus (get_current_line_type lexbuf))
          else (patch_or_match PATCH;
		add_current_line_type D.MINUS; token lexbuf) )
# 2915 "parsing_cocci/lexer_cocci.ml"

  | 14 ->
# 706 "parsing_cocci/lexer_cocci.mll"
        ( pass_zero();
	  if !current_line_started
	  then (start_line true; TPlus (get_current_line_type lexbuf))
          else (patch_or_match PATCH;
		add_current_line_type D.PLUS; token lexbuf) )
# 2924 "parsing_cocci/lexer_cocci.ml"

  | 15 ->
# 711 "parsing_cocci/lexer_cocci.mll"
        ( pass_zero();
	  if !current_line_started
	  then (start_line true; TWhy (get_current_line_type lexbuf))
          else (add_current_line_type D.OPT; token lexbuf) )
# 2932 "parsing_cocci/lexer_cocci.ml"

  | 16 ->
# 715 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TBang (get_current_line_type lexbuf) )
# 2937 "parsing_cocci/lexer_cocci.ml"

  | 17 ->
# 716 "parsing_cocci/lexer_cocci.mll"
        ( if not !col_zero
	  then (start_line true; TOPar (get_current_line_type lexbuf))
          else
            (start_line true; check_context_linetype (tok lexbuf);
	     TOPar0 ("(",get_current_line_type lexbuf)))
# 2946 "parsing_cocci/lexer_cocci.ml"

  | 18 ->
# 721 "parsing_cocci/lexer_cocci.mll"
          ( start_line true;
	    TOPar0 ("\\(",contextify(get_current_line_type lexbuf)) )
# 2952 "parsing_cocci/lexer_cocci.ml"

  | 19 ->
# 723 "parsing_cocci/lexer_cocci.mll"
        ( if not (!col_zero)
	  then (start_line true; TOr(get_current_line_type lexbuf))
          else (start_line true;
		check_context_linetype (tok lexbuf);
		TMid0 ("|",get_current_line_type lexbuf)))
# 2961 "parsing_cocci/lexer_cocci.ml"

  | 20 ->
# 728 "parsing_cocci/lexer_cocci.mll"
          ( start_line true;
	    TMid0 ("\\|",contextify(get_current_line_type lexbuf)) )
# 2967 "parsing_cocci/lexer_cocci.ml"

  | 21 ->
# 730 "parsing_cocci/lexer_cocci.mll"
        ( if not !col_zero
	  then (start_line true; TCPar (get_current_line_type lexbuf))
          else
            (start_line true; check_context_linetype (tok lexbuf);
	     TCPar0 (")",get_current_line_type lexbuf)))
# 2976 "parsing_cocci/lexer_cocci.ml"

  | 22 ->
# 735 "parsing_cocci/lexer_cocci.mll"
          ( start_line true;
	    TCPar0 ("\\)",contextify(get_current_line_type lexbuf)) )
# 2982 "parsing_cocci/lexer_cocci.ml"

  | 23 ->
# 738 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOCro (get_current_line_type lexbuf)   )
# 2987 "parsing_cocci/lexer_cocci.ml"

  | 24 ->
# 739 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCCro (get_current_line_type lexbuf)   )
# 2992 "parsing_cocci/lexer_cocci.ml"

  | 25 ->
# 740 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOBrace (get_current_line_type lexbuf) )
# 2997 "parsing_cocci/lexer_cocci.ml"

  | 26 ->
# 741 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCBrace (get_current_line_type lexbuf) )
# 3002 "parsing_cocci/lexer_cocci.ml"

  | 27 ->
# 743 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TPtrOp (get_current_line_type lexbuf)  )
# 3007 "parsing_cocci/lexer_cocci.ml"

  | 28 ->
# 744 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TDot (get_current_line_type lexbuf)    )
# 3012 "parsing_cocci/lexer_cocci.ml"

  | 29 ->
# 745 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TComma (get_current_line_type lexbuf)  )
# 3017 "parsing_cocci/lexer_cocci.ml"

  | 30 ->
# 746 "parsing_cocci/lexer_cocci.mll"
               ( start_line true; TVAEllipsis (get_current_line_type lexbuf) )
# 3022 "parsing_cocci/lexer_cocci.ml"

  | 31 ->
# 747 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TPtVirg (get_current_line_type lexbuf) )
# 3027 "parsing_cocci/lexer_cocci.ml"

  | 32 ->
# 750 "parsing_cocci/lexer_cocci.mll"
                   ( pass_zero();
		     if !current_line_started
		     then
		       (start_line true; TMul (get_current_line_type lexbuf))
		     else
		       (patch_or_match MATCH;
			add_current_line_type D.MINUS; token lexbuf) )
# 3038 "parsing_cocci/lexer_cocci.ml"

  | 33 ->
# 757 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TDmOp (Ast.Div,get_current_line_type lexbuf) )
# 3044 "parsing_cocci/lexer_cocci.ml"

  | 34 ->
# 759 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true;
		     TDmOp (Ast.Min,get_current_line_type lexbuf) )
# 3050 "parsing_cocci/lexer_cocci.ml"

  | 35 ->
# 761 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true;
		     TDmOp (Ast.Max,get_current_line_type lexbuf) )
# 3056 "parsing_cocci/lexer_cocci.ml"

  | 36 ->
# 763 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TDmOp (Ast.Mod,get_current_line_type lexbuf) )
# 3062 "parsing_cocci/lexer_cocci.ml"

  | 37 ->
# 765 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;  TTilde (get_current_line_type lexbuf) )
# 3067 "parsing_cocci/lexer_cocci.ml"

  | 38 ->
# 767 "parsing_cocci/lexer_cocci.mll"
                   ( pass_zero();
 		     if !current_line_started
 		     then
 		       (start_line true; TInc (get_current_line_type lexbuf))
 		     else (patch_or_match PATCH;
 			   add_current_line_type D.PLUSPLUS; token lexbuf) )
# 3077 "parsing_cocci/lexer_cocci.ml"

  | 39 ->
# 773 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;  TDec (get_current_line_type lexbuf) )
# 3082 "parsing_cocci/lexer_cocci.ml"

  | 40 ->
# 775 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TEq (get_current_line_type lexbuf) )
# 3087 "parsing_cocci/lexer_cocci.ml"

  | 41 ->
# 777 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Minus lexbuf )
# 3092 "parsing_cocci/lexer_cocci.ml"

  | 42 ->
# 778 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Plus lexbuf )
# 3097 "parsing_cocci/lexer_cocci.ml"

  | 43 ->
# 780 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Mul lexbuf )
# 3102 "parsing_cocci/lexer_cocci.ml"

  | 44 ->
# 781 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Div lexbuf )
# 3107 "parsing_cocci/lexer_cocci.ml"

  | 45 ->
# 782 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Mod lexbuf )
# 3112 "parsing_cocci/lexer_cocci.ml"

  | 46 ->
# 784 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.And lexbuf )
# 3117 "parsing_cocci/lexer_cocci.ml"

  | 47 ->
# 785 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Or lexbuf )
# 3122 "parsing_cocci/lexer_cocci.ml"

  | 48 ->
# 786 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Xor lexbuf )
# 3127 "parsing_cocci/lexer_cocci.ml"

  | 49 ->
# 787 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true; mkassign Ast.Max lexbuf )
# 3132 "parsing_cocci/lexer_cocci.ml"

  | 50 ->
# 788 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true; mkassign Ast.Min lexbuf )
# 3137 "parsing_cocci/lexer_cocci.ml"

  | 51 ->
# 790 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.DecLeft lexbuf )
# 3142 "parsing_cocci/lexer_cocci.ml"

  | 52 ->
# 791 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.DecRight lexbuf )
# 3147 "parsing_cocci/lexer_cocci.ml"

  | 53 ->
# 793 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TDotDot (get_current_line_type lexbuf) )
# 3152 "parsing_cocci/lexer_cocci.ml"

  | 54 ->
# 795 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TEqEq    (get_current_line_type lexbuf) )
# 3157 "parsing_cocci/lexer_cocci.ml"

  | 55 ->
# 796 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TNotEq   (get_current_line_type lexbuf) )
# 3162 "parsing_cocci/lexer_cocci.ml"

  | 56 ->
# 797 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TLogOp(Ast.SupEq,get_current_line_type lexbuf) )
# 3168 "parsing_cocci/lexer_cocci.ml"

  | 57 ->
# 799 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TLogOp(Ast.InfEq,get_current_line_type lexbuf) )
# 3174 "parsing_cocci/lexer_cocci.ml"

  | 58 ->
# 801 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TLogOp(Ast.Inf,get_current_line_type lexbuf) )
# 3180 "parsing_cocci/lexer_cocci.ml"

  | 59 ->
# 803 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TLogOp(Ast.Sup,get_current_line_type lexbuf) )
# 3186 "parsing_cocci/lexer_cocci.ml"

  | 60 ->
# 806 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TAndLog (get_current_line_type lexbuf) )
# 3191 "parsing_cocci/lexer_cocci.ml"

  | 61 ->
# 807 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TOrLog  (get_current_line_type lexbuf) )
# 3196 "parsing_cocci/lexer_cocci.ml"

  | 62 ->
# 809 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TShROp(Ast.DecRight,get_current_line_type lexbuf) )
# 3202 "parsing_cocci/lexer_cocci.ml"

  | 63 ->
# 811 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TShLOp(Ast.DecLeft,get_current_line_type lexbuf) )
# 3208 "parsing_cocci/lexer_cocci.ml"

  | 64 ->
# 814 "parsing_cocci/lexer_cocci.mll"
                   ( if not (!col_zero)
                     then (start_line true; TAnd(get_current_line_type lexbuf))
                     else (start_line true;
	                   check_context_linetype (tok lexbuf);
		           TAnd0 ("&",get_current_line_type lexbuf)))
# 3217 "parsing_cocci/lexer_cocci.ml"

  | 65 ->
# 819 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
	             TAnd0 ("\\&",contextify(get_current_line_type lexbuf)) )
# 3223 "parsing_cocci/lexer_cocci.ml"

  | 66 ->
# 822 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TXor(get_current_line_type lexbuf) )
# 3228 "parsing_cocci/lexer_cocci.ml"

  | 67 ->
# 824 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true; TCppConcatOp )
# 3233 "parsing_cocci/lexer_cocci.ml"

  | 68 ->
let
# 825 "parsing_cocci/lexer_cocci.mll"
                                               wss
# 3239 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 825 "parsing_cocci/lexer_cocci.mll"
                                                        def
# 3244 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(1)
and
# 826 "parsing_cocci/lexer_cocci.mll"
                                  ident
# 3249 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) lexbuf.Lexing.lex_curr_pos in
# 827 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let (arity,line,lline,llend,offset,col,strbef,straft,pos,ws) as lt =
	  get_current_line_type lexbuf in
	let off = String.length def in
	(* -1 in the code below because the ident is not at the line start *)
	TUndef
	  (lt,
	   check_var ident
	     (arity,line,lline,llend,offset+off,col+off,[],[],[],wss)) )
# 3261 "parsing_cocci/lexer_cocci.ml"

  | 69 ->
let
# 836 "parsing_cocci/lexer_cocci.mll"
                                                  wss
# 3267 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 836 "parsing_cocci/lexer_cocci.mll"
                                                            def
# 3272 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(1)
and
# 837 "parsing_cocci/lexer_cocci.mll"
                                   ident
# 3277 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) lexbuf.Lexing.lex_curr_pos in
# 838 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let (arity,line,lline,llend,offset,col,strbef,straft,pos,ws) as lt =
	  get_current_line_type lexbuf in
	let off = String.length def in
	(* -1 in the code below because the ident is not at the line start *)
	TDefine
	  (lt,
	   check_var ident
	     (arity,line,lline,llend,offset+off,col+off,[],[],[],wss)) )
# 3289 "parsing_cocci/lexer_cocci.ml"

  | 70 ->
let
# 847 "parsing_cocci/lexer_cocci.mll"
                                                  wss
# 3295 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 847 "parsing_cocci/lexer_cocci.mll"
                                                            def
# 3300 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(1)
and
# 848 "parsing_cocci/lexer_cocci.mll"
                                    ident
# 3305 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 850 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let (arity,line,lline,llend,offset,col,strbef,straft,pos,ws) as lt =
	  get_current_line_type lexbuf in
	let off = String.length def in
	TDefineParam
        (lt,
	 check_var ident
	   (* why pos here but not above? *)
	   (arity,line,lline,llend,offset+off,col+off,strbef,straft,pos,wss),
	 offset + off + (String.length ident),
	 col + off + (String.length ident)) )
# 3319 "parsing_cocci/lexer_cocci.ml"

  | 71 ->
let
# 861 "parsing_cocci/lexer_cocci.mll"
                                    prag
# 3325 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(0)
and
# 861 "parsing_cocci/lexer_cocci.mll"
                                                          wss1
# 3330 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 862 "parsing_cocci/lexer_cocci.mll"
                                    ident
# 3335 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) lexbuf.Lexing.lex_mem.(2)
and
# 862 "parsing_cocci/lexer_cocci.mll"
                                                           wss2
# 3340 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(3)
and
# 862 "parsing_cocci/lexer_cocci.mll"
                                                                               rest
# 3345 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(3) lexbuf.Lexing.lex_curr_pos in
# 863 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let (arity,line,lline,llend,offset,col,strbef,straft,pos,ws) as lt =
	  get_current_line_type lexbuf in
	let off1 = String.length prag + String.length wss1 in
	let off2 = off1 + String.length ident + String.length wss2 in
	TPragma(lt,
		check_var ident
	        (* why pos here but not above? *)
		  (arity,line,lline,llend,offset+off1,col+off1,strbef,straft,
		   pos,wss1),
		rest,
		(arity,line,lline,llend,offset+off2,col+off2,strbef,straft,
		 pos,wss2)) )
# 3361 "parsing_cocci/lexer_cocci.ml"

  | 72 ->
# 879 "parsing_cocci/lexer_cocci.mll"
      ( TIncludeL
	  (let str = tok lexbuf in
	  let start = String.index str '\"' in
	  let finish = String.rindex str '\"' in
	  start_line true;
	  (process_include start finish str, get_current_line_type lexbuf)) )
# 3371 "parsing_cocci/lexer_cocci.ml"

  | 73 ->
# 886 "parsing_cocci/lexer_cocci.mll"
      ( TIncludeNL
	  (let str = tok lexbuf in
	  let start = String.index str '<' in
	  let finish = String.rindex str '>' in
	  start_line true;
	  (process_include start finish str,get_current_line_type lexbuf)) )
# 3381 "parsing_cocci/lexer_cocci.ml"

  | 74 ->
# 893 "parsing_cocci/lexer_cocci.mll"
      ( check_minus_context_linetype (tok lexbuf);
	start_line true;
	TIncludeAny("...",get_current_line_type lexbuf) )
# 3388 "parsing_cocci/lexer_cocci.ml"

  | 75 ->
# 897 "parsing_cocci/lexer_cocci.mll"
      ( check_context_linetype (tok lexbuf);
	start_line true;
	TInclude(get_current_line_type lexbuf) )
# 3395 "parsing_cocci/lexer_cocci.ml"

  | 76 ->
# 908 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; check_plus_linetype (tok lexbuf);
	TDirective (Ast.Noindent(tok lexbuf), get_current_line_type lexbuf) )
# 3401 "parsing_cocci/lexer_cocci.ml"

  | 77 ->
# 912 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let res = TCppEscapedNewline (get_current_line_type lexbuf) in
	reset_line lexbuf;
	res )
# 3409 "parsing_cocci/lexer_cocci.ml"

  | 78 ->
# 918 "parsing_cocci/lexer_cocci.mll"
      (
       match !current_line_type with
        (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
        start_line true;
	(* second argument to TDirective is not quite right, because
	   it represents only the first token of the comment, but that
	   should be good enough *)
	TDirective (Ast.Indent("/*"^(comment check_comment lexbuf)),
		 get_current_line_type lexbuf)
      |	_ -> let _ = comment (fun _ -> ()) lexbuf in token lexbuf )
# 3423 "parsing_cocci/lexer_cocci.ml"

  | 79 ->
# 929 "parsing_cocci/lexer_cocci.mll"
      ( (if !current_line_started
      then lexerr "--- must be at the beginning of the line" "");
	start_line true;
	TMinusFile
	  (let str = tok lexbuf in
	  (drop_spaces(String.sub str 3 (String.length str - 3)),
	   (get_current_line_type lexbuf))) )
# 3434 "parsing_cocci/lexer_cocci.ml"

  | 80 ->
# 937 "parsing_cocci/lexer_cocci.mll"
      ( (if !current_line_started
      then lexerr "+++ must be at the beginning of the line" "");
	start_line true;
	TPlusFile
	  (let str = tok lexbuf in
	  (drop_spaces(String.sub str 3 (String.length str - 3)),
	   (get_current_line_type lexbuf))) )
# 3445 "parsing_cocci/lexer_cocci.ml"

  | 81 ->
# 946 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; id_tokens lexbuf )
# 3450 "parsing_cocci/lexer_cocci.ml"

  | 82 ->
# 950 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; id_tokens lexbuf )
# 3455 "parsing_cocci/lexer_cocci.ml"

  | 83 ->
# 957 "parsing_cocci/lexer_cocci.mll"
      (
	start_line true;
	if !Flag.c_plus_plus = Flag.Off
	then
	  Common.pr2_once
	    "< and > not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3467 "parsing_cocci/lexer_cocci.ml"

  | 84 ->
# 968 "parsing_cocci/lexer_cocci.mll"
      (
	start_line true;
	if !Flag.c_plus_plus = Flag.Off
	then
	  Common.pr2_once
	    "< and > not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3479 "parsing_cocci/lexer_cocci.ml"

  | 85 ->
# 984 "parsing_cocci/lexer_cocci.mll"
      (
	start_line true;
	if !Flag.c_plus_plus = Flag.Off
	then
	  Common.pr2_once
	    "~ and :: not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3491 "parsing_cocci/lexer_cocci.ml"

  | 86 ->
# 997 "parsing_cocci/lexer_cocci.mll"
      (
	start_line true;
	if !Flag.c_plus_plus = Flag.Off
	then
	  Common.pr2_once
	    "~ and :: not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3503 "parsing_cocci/lexer_cocci.ml"

  | 87 ->
# 1008 "parsing_cocci/lexer_cocci.mll"
        ( start_line true;
	  TChar(char lexbuf,Ast.IsChar,get_current_line_type lexbuf) )
# 3509 "parsing_cocci/lexer_cocci.ml"

  | 88 ->
# 1010 "parsing_cocci/lexer_cocci.mll"
         ( start_line true;
	  TString(string lexbuf,Ast.IsChar,(get_current_line_type lexbuf)) )
# 3515 "parsing_cocci/lexer_cocci.ml"

  | 89 ->
# 1012 "parsing_cocci/lexer_cocci.mll"
            ( start_line true;
	  TChar(char lexbuf,Ast.IsWchar,get_current_line_type lexbuf) )
# 3521 "parsing_cocci/lexer_cocci.ml"

  | 90 ->
# 1014 "parsing_cocci/lexer_cocci.mll"
             ( start_line true;
	  TString(string lexbuf,Ast.IsWchar,(get_current_line_type lexbuf)) )
# 3527 "parsing_cocci/lexer_cocci.ml"

  | 91 ->
# 1016 "parsing_cocci/lexer_cocci.mll"
            ( start_line true;
	  TChar(char lexbuf,Ast.IsUchar,get_current_line_type lexbuf) )
# 3533 "parsing_cocci/lexer_cocci.ml"

  | 92 ->
# 1018 "parsing_cocci/lexer_cocci.mll"
             ( start_line true;
	  TString(string lexbuf,Ast.IsUchar,(get_current_line_type lexbuf)) )
# 3539 "parsing_cocci/lexer_cocci.ml"

  | 93 ->
# 1020 "parsing_cocci/lexer_cocci.mll"
            ( start_line true;
	  TChar(char lexbuf,Ast.Isuchar,get_current_line_type lexbuf) )
# 3545 "parsing_cocci/lexer_cocci.ml"

  | 94 ->
# 1022 "parsing_cocci/lexer_cocci.mll"
             ( start_line true;
	  TString(string lexbuf,Ast.Isuchar,(get_current_line_type lexbuf)) )
# 3551 "parsing_cocci/lexer_cocci.ml"

  | 95 ->
# 1024 "parsing_cocci/lexer_cocci.mll"
              ( start_line true;
	  TString(string lexbuf,Ast.Isu8char,(get_current_line_type lexbuf)) )
# 3557 "parsing_cocci/lexer_cocci.ml"

  | 96 ->
let
# 1026 "parsing_cocci/lexer_cocci.mll"
             x
# 3563 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1026 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TFloat(x,(get_current_line_type lexbuf)) )
# 3568 "parsing_cocci/lexer_cocci.ml"

  | 97 ->
let
# 1037 "parsing_cocci/lexer_cocci.mll"
         x
# 3574 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1037 "parsing_cocci/lexer_cocci.mll"
            ( start_line true; TInt(x,(get_current_line_type lexbuf)) )
# 3578 "parsing_cocci/lexer_cocci.ml"

  | 98 ->
let
# 1039 "parsing_cocci/lexer_cocci.mll"
                           x
# 3584 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1040 "parsing_cocci/lexer_cocci.mll"
      ( if !Flag.ibm
      then
	begin
	  start_line true;
	  let len = string_of_int(String.length x - 1) in
          TDecimalCst(x,len,"0",(get_current_line_type lexbuf))
	end
      else failwith "unrecognized constant modifier d/D" )
# 3595 "parsing_cocci/lexer_cocci.ml"

  | 99 ->
# 1049 "parsing_cocci/lexer_cocci.mll"
                   ( TIso )
# 3600 "parsing_cocci/lexer_cocci.ml"

  | 100 ->
# 1050 "parsing_cocci/lexer_cocci.mll"
                   ( TRightIso )
# 3605 "parsing_cocci/lexer_cocci.ml"

  | 101 ->
# 1052 "parsing_cocci/lexer_cocci.mll"
                   ( EOF )
# 3610 "parsing_cocci/lexer_cocci.ml"

  | 102 ->
# 1054 "parsing_cocci/lexer_cocci.mll"
      ( lexerr "unrecognised symbol, in token rule: " (tok lexbuf) )
# 3615 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_token_rec lexbuf __ocaml_lex_state

and metavariable_decl_token lexbuf =
   __ocaml_lex_metavariable_decl_token_rec lexbuf 242
and __ocaml_lex_metavariable_decl_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1059 "parsing_cocci/lexer_cocci.mll"
    ( reset_line lexbuf; metavariable_decl_token lexbuf )
# 3627 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
# 1061 "parsing_cocci/lexer_cocci.mll"
                  (
    start_line false; metavariable_decl_token lexbuf )
# 3633 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
# 1064 "parsing_cocci/lexer_cocci.mll"
                                   (
    start_line false; metavariable_decl_token lexbuf )
# 3639 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
# 1067 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TArobArob )
# 3644 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
# 1069 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TTildeEq (get_current_line_type lexbuf) )
# 3649 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
# 1070 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TTildeExclEq (get_current_line_type lexbuf) )
# 3654 "parsing_cocci/lexer_cocci.ml"

  | 6 ->
# 1071 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TEq (get_current_line_type lexbuf) )
# 3659 "parsing_cocci/lexer_cocci.ml"

  | 7 ->
# 1072 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOPar (get_current_line_type lexbuf) )
# 3664 "parsing_cocci/lexer_cocci.ml"

  | 8 ->
# 1073 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCPar (get_current_line_type lexbuf) )
# 3669 "parsing_cocci/lexer_cocci.ml"

  | 9 ->
# 1075 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOCro (get_current_line_type lexbuf)   )
# 3674 "parsing_cocci/lexer_cocci.ml"

  | 10 ->
# 1076 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCCro (get_current_line_type lexbuf)   )
# 3679 "parsing_cocci/lexer_cocci.ml"

  | 11 ->
# 1077 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOBrace (get_current_line_type lexbuf) )
# 3684 "parsing_cocci/lexer_cocci.ml"

  | 12 ->
# 1078 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCBrace (get_current_line_type lexbuf) )
# 3689 "parsing_cocci/lexer_cocci.ml"

  | 13 ->
# 1080 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TPtrOp (get_current_line_type lexbuf)  )
# 3694 "parsing_cocci/lexer_cocci.ml"

  | 14 ->
# 1081 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TDot (get_current_line_type lexbuf)    )
# 3699 "parsing_cocci/lexer_cocci.ml"

  | 15 ->
# 1082 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TComma (get_current_line_type lexbuf)  )
# 3704 "parsing_cocci/lexer_cocci.ml"

  | 16 ->
# 1083 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TMPtVirg (* works better with tokens_all *) )
# 3710 "parsing_cocci/lexer_cocci.ml"

  | 17 ->
# 1085 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TShLOp(Ast.DecLeft,get_current_line_type lexbuf) )
# 3716 "parsing_cocci/lexer_cocci.ml"

  | 18 ->
# 1088 "parsing_cocci/lexer_cocci.mll"
                   ( pass_zero();
		     if !current_line_started
		     then
		       (start_line true; TMul (get_current_line_type lexbuf))
		     else
		       (patch_or_match MATCH;
			add_current_line_type D.MINUS;
			metavariable_decl_token lexbuf) )
# 3728 "parsing_cocci/lexer_cocci.ml"

  | 19 ->
# 1097 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TEqEq    (get_current_line_type lexbuf) )
# 3733 "parsing_cocci/lexer_cocci.ml"

  | 20 ->
# 1098 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TNotEq   (get_current_line_type lexbuf) )
# 3738 "parsing_cocci/lexer_cocci.ml"

  | 21 ->
# 1099 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TSub     (get_current_line_type lexbuf) )
# 3743 "parsing_cocci/lexer_cocci.ml"

  | 22 ->
# 1100 "parsing_cocci/lexer_cocci.mll"
        ( (start_line true; TPlus (get_current_line_type lexbuf)) )
# 3748 "parsing_cocci/lexer_cocci.ml"

  | 23 ->
# 1101 "parsing_cocci/lexer_cocci.mll"
        ( (start_line true; TMinus (get_current_line_type lexbuf)) )
# 3753 "parsing_cocci/lexer_cocci.ml"

  | 24 ->
# 1102 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TDmOp (Ast.Div,get_current_line_type lexbuf) )
# 3758 "parsing_cocci/lexer_cocci.ml"

  | 25 ->
# 1103 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TDmOp (Ast.Mod,get_current_line_type lexbuf) )
# 3763 "parsing_cocci/lexer_cocci.ml"

  | 26 ->
# 1104 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TShROp(Ast.DecRight,get_current_line_type lexbuf) )
# 3768 "parsing_cocci/lexer_cocci.ml"

  | 27 ->
# 1105 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TAnd (get_current_line_type lexbuf) )
# 3773 "parsing_cocci/lexer_cocci.ml"

  | 28 ->
# 1106 "parsing_cocci/lexer_cocci.mll"
        (  (start_line true; TOr(get_current_line_type lexbuf)) )
# 3778 "parsing_cocci/lexer_cocci.ml"

  | 29 ->
# 1107 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TXor(get_current_line_type lexbuf) )
# 3783 "parsing_cocci/lexer_cocci.ml"

  | 30 ->
# 1108 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TLogOp(Ast.SupEq,get_current_line_type lexbuf) )
# 3788 "parsing_cocci/lexer_cocci.ml"

  | 31 ->
# 1109 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TLogOp(Ast.Inf,get_current_line_type lexbuf) )
# 3793 "parsing_cocci/lexer_cocci.ml"

  | 32 ->
# 1110 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TLogOp(Ast.Sup,get_current_line_type lexbuf) )
# 3798 "parsing_cocci/lexer_cocci.ml"

  | 33 ->
# 1111 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TAndLog (get_current_line_type lexbuf) )
# 3803 "parsing_cocci/lexer_cocci.ml"

  | 34 ->
# 1112 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TOrLog (get_current_line_type lexbuf) )
# 3808 "parsing_cocci/lexer_cocci.ml"

  | 35 ->
# 1113 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TDotDot (get_current_line_type lexbuf) )
# 3813 "parsing_cocci/lexer_cocci.ml"

  | 36 ->
# 1114 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Minus lexbuf )
# 3818 "parsing_cocci/lexer_cocci.ml"

  | 37 ->
# 1115 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Plus lexbuf )
# 3823 "parsing_cocci/lexer_cocci.ml"

  | 38 ->
# 1116 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Mul lexbuf )
# 3828 "parsing_cocci/lexer_cocci.ml"

  | 39 ->
# 1117 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Div lexbuf )
# 3833 "parsing_cocci/lexer_cocci.ml"

  | 40 ->
# 1118 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Mod lexbuf )
# 3838 "parsing_cocci/lexer_cocci.ml"

  | 41 ->
# 1119 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.And lexbuf )
# 3843 "parsing_cocci/lexer_cocci.ml"

  | 42 ->
# 1120 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Or lexbuf )
# 3848 "parsing_cocci/lexer_cocci.ml"

  | 43 ->
# 1121 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Xor lexbuf )
# 3853 "parsing_cocci/lexer_cocci.ml"

  | 44 ->
# 1122 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.DecLeft lexbuf )
# 3858 "parsing_cocci/lexer_cocci.ml"

  | 45 ->
# 1123 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.DecRight lexbuf )
# 3863 "parsing_cocci/lexer_cocci.ml"

  | 46 ->
# 1125 "parsing_cocci/lexer_cocci.mll"
      (match !current_line_type with
        (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
        start_line true;
	(* second argument to TDirective is not quite right, because
	   it represents only the first token of the comment, but that
	   should be good enough *)
	TDirective (Ast.Indent("/*"^(comment check_comment lexbuf)),
		 get_current_line_type lexbuf)
      |	_ -> let _ = comment (fun _ -> ()) lexbuf in
	     metavariable_decl_token lexbuf )
# 3877 "parsing_cocci/lexer_cocci.ml"

  | 47 ->
# 1136 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true; TCppConcatOp (* for fresh vars *) )
# 3882 "parsing_cocci/lexer_cocci.ml"

  | 48 ->
# 1139 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; id_tokens lexbuf )
# 3887 "parsing_cocci/lexer_cocci.ml"

  | 49 ->
# 1143 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; id_tokens lexbuf )
# 3892 "parsing_cocci/lexer_cocci.ml"

  | 50 ->
# 1150 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	if !Flag.c_plus_plus = Flag.Off
	then
	  Common.pr2_once
	    "< and > not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3903 "parsing_cocci/lexer_cocci.ml"

  | 51 ->
# 1160 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	if !Flag.c_plus_plus = Flag.Off
	then
	  Common.pr2_once
	    "< and > not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3914 "parsing_cocci/lexer_cocci.ml"

  | 52 ->
# 1175 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	if !Flag.c_plus_plus = Flag.Off
	then
	  Common.pr2_once
	    "~ and :: not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3925 "parsing_cocci/lexer_cocci.ml"

  | 53 ->
# 1187 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	if !Flag.c_plus_plus = Flag.Off
	then
	  Common.pr2_once
	    "~ and :: not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3936 "parsing_cocci/lexer_cocci.ml"

  | 54 ->
# 1197 "parsing_cocci/lexer_cocci.mll"
        ( start_line true;
	  TChar(char lexbuf,Ast.IsChar,get_current_line_type lexbuf) )
# 3942 "parsing_cocci/lexer_cocci.ml"

  | 55 ->
# 1199 "parsing_cocci/lexer_cocci.mll"
         ( start_line true;
	  TString(string lexbuf,Ast.IsChar,(get_current_line_type lexbuf)) )
# 3948 "parsing_cocci/lexer_cocci.ml"

  | 56 ->
# 1201 "parsing_cocci/lexer_cocci.mll"
            ( start_line true;
	  TChar(char lexbuf,Ast.IsWchar,get_current_line_type lexbuf) )
# 3954 "parsing_cocci/lexer_cocci.ml"

  | 57 ->
# 1203 "parsing_cocci/lexer_cocci.mll"
             ( start_line true;
	  TString(string lexbuf,Ast.IsWchar,(get_current_line_type lexbuf)) )
# 3960 "parsing_cocci/lexer_cocci.ml"

  | 58 ->
# 1205 "parsing_cocci/lexer_cocci.mll"
            ( start_line true;
	  TChar(char lexbuf,Ast.IsUchar,get_current_line_type lexbuf) )
# 3966 "parsing_cocci/lexer_cocci.ml"

  | 59 ->
# 1207 "parsing_cocci/lexer_cocci.mll"
             ( start_line true;
	  TString(string lexbuf,Ast.IsUchar,(get_current_line_type lexbuf)) )
# 3972 "parsing_cocci/lexer_cocci.ml"

  | 60 ->
# 1209 "parsing_cocci/lexer_cocci.mll"
            ( start_line true;
	  TChar(char lexbuf,Ast.Isuchar,get_current_line_type lexbuf) )
# 3978 "parsing_cocci/lexer_cocci.ml"

  | 61 ->
# 1211 "parsing_cocci/lexer_cocci.mll"
             ( start_line true;
	  TString(string lexbuf,Ast.Isuchar,(get_current_line_type lexbuf)) )
# 3984 "parsing_cocci/lexer_cocci.ml"

  | 62 ->
# 1213 "parsing_cocci/lexer_cocci.mll"
              ( start_line true;
	  TString(string lexbuf,Ast.Isu8char,(get_current_line_type lexbuf)) )
# 3990 "parsing_cocci/lexer_cocci.ml"

  | 63 ->
let
# 1215 "parsing_cocci/lexer_cocci.mll"
             x
# 3996 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1215 "parsing_cocci/lexer_cocci.mll"
                   ( TFloat(x,(get_current_line_type lexbuf)) )
# 4000 "parsing_cocci/lexer_cocci.ml"

  | 64 ->
let
# 1224 "parsing_cocci/lexer_cocci.mll"
         x
# 4006 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1224 "parsing_cocci/lexer_cocci.mll"
            ( start_line true; TInt(x,(get_current_line_type lexbuf)) )
# 4010 "parsing_cocci/lexer_cocci.ml"

  | 65 ->
let
# 1226 "parsing_cocci/lexer_cocci.mll"
                           x
# 4016 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1227 "parsing_cocci/lexer_cocci.mll"
      ( if !Flag.ibm
      then
	begin
	  start_line true;
	  let len = string_of_int(String.length x - 1) in
          TDecimalCst(x,len,"0",(get_current_line_type lexbuf))
	end
      else failwith "unrecognized constant modifier d/D" )
# 4027 "parsing_cocci/lexer_cocci.ml"

  | 66 ->
# 1236 "parsing_cocci/lexer_cocci.mll"
      ( lexerr
	 "metavariables: unrecognised symbol in metavariable_decl_token rule: "
	 (tok lexbuf) )
# 4034 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_metavariable_decl_token_rec lexbuf __ocaml_lex_state

and char lexbuf =
   __ocaml_lex_char_rec lexbuf 361
and __ocaml_lex_char_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 1242 "parsing_cocci/lexer_cocci.mll"
          x
# 4047 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 1242 "parsing_cocci/lexer_cocci.mll"
                                       ( String.make 1 x ^ restchars lexbuf )
# 4051 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
let
# 1244 "parsing_cocci/lexer_cocci.mll"
                                             x
# 4057 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1244 "parsing_cocci/lexer_cocci.mll"
                                                     ( x ^ restchars lexbuf )
# 4061 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
let
# 1248 "parsing_cocci/lexer_cocci.mll"
                                  x
# 4067 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1248 "parsing_cocci/lexer_cocci.mll"
                                                     ( x ^ restchars lexbuf )
# 4071 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
let
# 1249 "parsing_cocci/lexer_cocci.mll"
                 v
# 4077 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
and
# 1249 "parsing_cocci/lexer_cocci.mll"
                                  x
# 4082 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 1250 "parsing_cocci/lexer_cocci.mll"
 (
          (match v with (* Machine specific ? *)
          | 'n' -> ()  | 't' -> ()   | 'v' -> ()  | 'b' -> () | 'r' -> ()
          | 'f' -> () | 'a' -> ()
	  | '\\' -> () | '?'  -> () | '\'' -> ()  | '\"' -> ()
          | 'e' -> () (* linuxext: ? *)
	  | _ ->
              Common.pr2 ("LEXER: unrecognised symbol in char:"^tok lexbuf);
	  );
          x ^ restchars lexbuf
	)
# 4096 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
# 1262 "parsing_cocci/lexer_cocci.mll"
      ( Common.pr2 ("LEXER: unrecognised symbol in char:"^tok lexbuf);
        tok lexbuf ^ restchars lexbuf
      )
# 4103 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_char_rec lexbuf __ocaml_lex_state

and restchars lexbuf =
   __ocaml_lex_restchars_rec lexbuf 371
and __ocaml_lex_restchars_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1267 "parsing_cocci/lexer_cocci.mll"
                                       ( "" )
# 4115 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
let
# 1268 "parsing_cocci/lexer_cocci.mll"
          x
# 4121 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 1268 "parsing_cocci/lexer_cocci.mll"
                                       ( String.make 1 x ^ restchars lexbuf )
# 4125 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
let
# 1270 "parsing_cocci/lexer_cocci.mll"
                                             x
# 4131 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1270 "parsing_cocci/lexer_cocci.mll"
                                                     ( x ^ restchars lexbuf )
# 4135 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
let
# 1274 "parsing_cocci/lexer_cocci.mll"
                                  x
# 4141 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1274 "parsing_cocci/lexer_cocci.mll"
                                                     ( x ^ restchars lexbuf )
# 4145 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
let
# 1275 "parsing_cocci/lexer_cocci.mll"
                 v
# 4151 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
and
# 1275 "parsing_cocci/lexer_cocci.mll"
                                  x
# 4156 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 1276 "parsing_cocci/lexer_cocci.mll"
 (
          (match v with (* Machine specific ? *)
          | 'n' -> ()  | 't' -> ()   | 'v' -> ()  | 'b' -> () | 'r' -> ()
          | 'f' -> () | 'a' -> ()
	  | '\\' -> () | '?'  -> () | '\'' -> ()  | '\"' -> ()
          | 'e' -> () (* linuxext: ? *)
	  | _ ->
              Common.pr2 ("LEXER: unrecognised symbol in char:"^tok lexbuf);
	  );
          x ^ restchars lexbuf
	)
# 4170 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
# 1288 "parsing_cocci/lexer_cocci.mll"
      ( Common.pr2 ("LEXER: unrecognised symbol in char:"^tok lexbuf);
        tok lexbuf ^ restchars lexbuf
      )
# 4177 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_restchars_rec lexbuf __ocaml_lex_state

and string lexbuf =
   __ocaml_lex_string_rec lexbuf 382
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1293 "parsing_cocci/lexer_cocci.mll"
                                               ( "" )
# 4189 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
let
# 1294 "parsing_cocci/lexer_cocci.mll"
                                 x
# 4195 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 1295 "parsing_cocci/lexer_cocci.mll"
    ( line := !line + 1; (Printf.sprintf "%c" x) ^ string lexbuf )
# 4199 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
let
# 1296 "parsing_cocci/lexer_cocci.mll"
          x
# 4205 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 1296 "parsing_cocci/lexer_cocci.mll"
                               ( Common.string_of_char x ^ string lexbuf )
# 4209 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
let
# 1297 "parsing_cocci/lexer_cocci.mll"
                                            x
# 4215 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1297 "parsing_cocci/lexer_cocci.mll"
                                              ( x ^ string lexbuf )
# 4219 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
let
# 1298 "parsing_cocci/lexer_cocci.mll"
                               x
# 4225 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1298 "parsing_cocci/lexer_cocci.mll"
                                              ( x ^ string lexbuf )
# 4229 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
let
# 1299 "parsing_cocci/lexer_cocci.mll"
                v
# 4235 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
and
# 1299 "parsing_cocci/lexer_cocci.mll"
                       x
# 4240 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 1300 "parsing_cocci/lexer_cocci.mll"
       (
         (match v with
	    | 'n' -> ()  | 't' -> ()   | 'v' -> ()  | 'b' -> () | 'r' -> ()
	    | 'f' -> () | 'a' -> ()
	    | '\\' -> () | '?'  -> () | '\'' -> ()  | '\"' -> ()
	    | 'e' -> ()
	    | '\n' -> ()
	    | '(' -> () | '|' -> () | ')' -> ()
	    | _ -> lexerr "unrecognised symbol:" (tok lexbuf)
	 );
          x ^ string lexbuf
       )
# 4255 "parsing_cocci/lexer_cocci.ml"

  | 6 ->
# 1312 "parsing_cocci/lexer_cocci.mll"
      ( lexerr "unrecognised symbol: " (tok lexbuf) )
# 4260 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

and comment check_comment lexbuf =
   __ocaml_lex_comment_rec check_comment lexbuf 394
and __ocaml_lex_comment_rec check_comment lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1315 "parsing_cocci/lexer_cocci.mll"
         ( let s = tok lexbuf in check_comment s; start_line true; s )
# 4272 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
# 1317 "parsing_cocci/lexer_cocci.mll"
      ( let s = tok lexbuf in
        (* even blank line should have a + *)
        check_comment s;
        reset_line lexbuf; s ^ comment check_comment lexbuf )
# 4280 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
# 1321 "parsing_cocci/lexer_cocci.mll"
        ( pass_zero();
	  if !current_line_started
	  then (start_line true;
		let s = tok lexbuf in s^(comment check_comment lexbuf))
	  else (start_line true; comment check_comment lexbuf) )
# 4289 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
# 1328 "parsing_cocci/lexer_cocci.mll"
      ( let s = tok lexbuf in
        check_comment s; start_line true; s ^ comment check_comment lexbuf )
# 4295 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
# 1331 "parsing_cocci/lexer_cocci.mll"
      ( let s = tok lexbuf in
        check_comment s; start_line true; s ^ comment check_comment lexbuf )
# 4301 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
# 1334 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; let s = tok lexbuf in
        Common.pr2 ("LEXER: unrecognised symbol in comment:"^s);
        s ^ comment check_comment lexbuf
      )
# 4309 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comment_rec check_comment lexbuf __ocaml_lex_state

;;

